# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper classes for managing issues."""

import re


def get_values_containing(target, expression):
  regex = re.compile(expression, re.DOTALL | re.IGNORECASE)
  return [value for value in target if regex.search(value)]


def get_values_matching(target, expression):
  regex = re.compile(expression + r'\Z', re.DOTALL | re.IGNORECASE)
  return [value for value in target if regex.match(value)]


def has_values_containing(target, expression):
  return any(get_values_containing(target, expression))


def has_values_matching(target, expression):
  return any(get_values_matching(target, expression))


def has_value(target, value):
  for target_value in target:
    if target_value.lower() == value.lower():
      return True

  return False


class ChangeList(list):
  """List that tracks changes for incremental updates."""

  def __init__(self, seq=()):
    super(ChangeList, self).__init__(seq)
    self.added = set()
    self.removed = set()

  def append(self, p_object):
    list.append(self, p_object)

    if p_object in self.removed:
      self.removed.remove(p_object)
    else:
      self.added.add(p_object)

  def remove(self, value):
    list.remove(self, value)

    if value in self.added:
      self.added.remove(value)
    else:
      self.removed.add(value)

  def is_changed(self):
    return (len(self.added) + len(self.removed)) > 0

  def reset(self):
    self.added.clear()
    self.removed.clear()


class Issue(object):
  """Class representing a single issue."""

  def __init__(self):
    self.blocking = None
    self.blocked_on = None
    self.body = None
    self.depends_on = None
    self.cc = ChangeList()
    self.closed = None
    self.comment = ''
    self.components = ChangeList()
    self.created = None
    self.id = 0
    self.labels = ChangeList()
    self.merged_into = None
    self.merged_into_project = None
    self.open = False
    self.owner = None
    self.reporter = None
    self.status = None
    self.stars = 0
    self.summary = None
    self.updated = None

    self.dirty = False
    self.send_email = True
    self.new = True
    self.itm = None
    self.project_name = None
    self.comments = None
    self.comment_count = 0
    self.first_comment = None
    self.last_comment = None
    self.changed = set()

  def __getattribute__(self, item):
    if item in ['body'] and not object.__getattribute__(self, item):
      comment = self.get_first_comment()
      self.__setattr__(item, comment.comment)

    return object.__getattribute__(self, item)

  def __setattr__(self, name, value):
    self.__dict__[name] = value
    if 'changed' in self.__dict__:
      self.__dict__['changed'].add(name)

    # Automatically set the project name if the itm is set.
    if name == 'itm' and value and hasattr(value, 'project_name'):
      self.__dict__['project_name'] = value.project_name

    # Treat comments and dirty flag specially.
    if name not in ('dirty', 'body', 'comments', 'itm', 'new', 'comment_count',
                    'first_comment', 'last_comment', 'project_name', 'changed',
                    'send_email'):
      self.__dict__['dirty'] = True

    if name in ('dirty') and not value:
      self.labels.reset()
      self.cc.reset()
      if 'changed' in self.__dict__:
        self.changed.clear()

  def __getstate__(self):
    """Ensure that we don't pickle the itm.

    This would raise an exception due to the way the apiary folks did their
    information (i.e. OAuth kicking us once again).
    """
    result_dict = self.__dict__.copy()
    del result_dict['itm']
    return result_dict

  def __setstate__(self, new_dict):
    self.__dict__.update(new_dict)
    self.itm = None

  def _remove_tracked_value(self, target, value):
    for existing_value in target:
      if existing_value.lower() == value.lower():
        target.remove(existing_value)
        self.dirty = True
        return

  def add_component(self, component):
    if not self.has_component(component):
      self.components.append(component)
      self.dirty = True

  def remove_component(self, component):
    if self.has_component(component):
      self._remove_tracked_value(self.components, component)
      self.add_component('-%s' % component)

  def remove_components_by_prefix(self, prefix):
    components = self.get_components_by_prefix(prefix)
    for component in components:
      self.remove_label(component)

  def add_label(self, label):
    if not self.has_label(label):
      self.labels.append(label)
      self.dirty = True

  def remove_label(self, label):
    if self.has_label(label):
      self._remove_tracked_value(self.labels, label)
      self.add_label('-%s' % label)

  def remove_label_by_prefix(self, prefix):
    labels = self.get_labels_by_prefix(prefix)
    for label in labels:
      self.remove_label(label)

  def add_cc(self, cc):
    if not self.has_cc(cc):
      self.cc.append(cc)
      self.dirty = True

  def remove_cc(self, cc):
    if self.has_cc(cc):
      self.cc.remove(cc)
      self.dirty = True

  def get_components_by_prefix(self, prefix):
    return get_values_matching(self.components, '%s.*' % prefix)

  def get_components_containing(self, expression):
    return get_values_containing(self.components, expression)

  def get_components_matching(self, expression):
    return get_values_matching(self.components, expression)

  def has_components_containing(self, expression):
    return has_values_containing(self.components, expression)

  def has_components_matching(self, expression):
    return has_values_matching(self.components, expression)

  def has_component(self, value):
    return has_value(self.components, value)

  def get_labels_by_prefix(self, prefix):
    return get_values_matching(self.labels, '%s.*' % prefix)

  def get_labels_containing(self, expression):
    return get_values_containing(self.labels, expression)

  def get_labels_matching(self, expression):
    return get_values_matching(self.labels, expression)

  def has_label_by_prefix(self, prefix):
    return has_values_containing(self.labels, '%s.*' % prefix)

  def has_label_containing(self, expression):
    return has_values_containing(self.labels, expression)

  def has_label_matching(self, expression):
    return has_values_matching(self.labels, expression)

  def has_label(self, value):
    return has_value(self.labels, value)

  def has_cc(self, value):
    return has_value(self.cc, value)

  def has_comment_with_label(self, label):
    for comment in self.get_comments():
      if comment.has_label(label):
        return True
    return False

  def has_comment_with_label_by_prefix(self, prefix):
    for comment in self.get_comments():
      if comment.get_labels_by_prefix(prefix):
        return True
    return False

  def get_comments(self):
    if not self.comments and self.itm:
      self.comments = self.itm.get_comments(self.id)
      self.comment_count = len(self.comments)
    return self.comments

  def get_first_comment(self):
    if not self.first_comment and self.itm:
      self.first_comment = self.itm.get_first_comment(self.id)
    return self.first_comment

  def get_last_comment(self):
    if not self.last_comment and self.itm:
      self.last_comment = self.itm.get_last_comment(self.id)
    return self.last_comment

  def get_comment_count(self):
    if not self.comment_count and self.itm:
      self.comment_count = self.itm.get_comment_count(self.id)
    return self.comment_count

  def save(self, send_email=None):
    if self.itm:
      self.itm.save(self, send_email)

  def refresh(self):
    if self.itm:
      self.comments = None
      self.last_comment = None
      self.comment_count = 0
      self.itm.refresh(self)
    return self



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Login page."""

import datetime

from flask import request

from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import auth
from libs import handler
from libs import helpers

DEFAULT_REDIRECT = '/'
SESSION_EXPIRY_DAYS = 14


class Handler(base_handler.Handler):
  """Login page."""

  @handler.get(handler.HTML)
  @handler.unsupported_on_local_server
  def get(self):
    """Handle a get request."""
    dest = request.get('dest', DEFAULT_REDIRECT)
    base_handler.check_redirect_url(dest)

    return self.render(
        'login.html', {
            'apiKey': local_config.ProjectConfig().get('firebase.api_key'),
            'authDomain': auth.auth_domain(),
            'dest': dest,
        })


class SessionLoginHandler(base_handler.Handler):
  """Session login handler."""

  @handler.post(handler.JSON, handler.JSON)
  def post(self):
    """Handle a post request."""
    id_token = request.get('idToken')
    expires_in = datetime.timedelta(days=SESSION_EXPIRY_DAYS)
    try:
      session_cookie = auth.create_session_cookie(id_token, expires_in)
    except auth.AuthError:
      raise helpers.EarlyExitException('Failed to create session cookie.', 401)

    expires = datetime.datetime.now() + expires_in
    response = self.render_json({'status': 'success'})
    response.set_cookie(
        'session', session_cookie, expires=expires, httponly=True, secure=True)
    return response


class LogoutHandler(base_handler.Handler):
  """Log out handler."""

  @handler.get(handler.HTML)
  @handler.unsupported_on_local_server
  @handler.require_csrf_token
  def get(self):
    """Handle a get request."""
    try:
      auth.revoke_session_cookie(auth.get_session_cookie())
    except auth.AuthError:
      # Even if the revoke failed, remove the cookie.
      logs.log_error('Failed to revoke session cookie.')

    response = self.redirect(request.get('dest', DEFAULT_REDIRECT))
    response.delete_cookie('session')
    return response



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""run.py runs a one-off script (e.g. migration) with appropriate envs (e.g.
  datastore)."""

import importlib
import os

from local.butler import constants
from src.clusterfuzz._internal.config import local_config
from src.clusterfuzz._internal.datastore import ndb_init


def execute(args):
  """Run Python unit tests under v2. For unittests involved appengine, sys.path
     needs certain modification."""
  os.environ['CONFIG_DIR_OVERRIDE'] = args.config_dir
  local_config.ProjectConfig().set_environment()

  if args.local:
    os.environ['DATASTORE_EMULATOR_HOST'] = constants.DATASTORE_EMULATOR_HOST
    os.environ['PUBSUB_EMULATOR_HOST'] = constants.PUBSUB_EMULATOR_HOST
    os.environ['DATASTORE_USE_PROJECT_ID_AS_APP_ID'] = 'true'
    os.environ['LOCAL_DEVELOPMENT'] = 'True'

  if not args.non_dry_run:
    print('Running in dry-run mode, no datastore writes are committed. '
          'For permanent modifications, re-run with --non-dry-run.')

  with ndb_init.context():
    script = importlib.import_module(
        'local.butler.scripts.%s' % args.script_name)
    script.execute(args)

  if not args.local:
    print()
    print('Please remember to run the migration individually on all projects.')
    print()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Crash minidump and symbols uploader."""

import email
import os
import re
import tempfile

import requests

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_parser
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms.android import adb
from clusterfuzz._internal.platforms.android import constants
from clusterfuzz._internal.protos import process_state_pb2
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler

CRASH_REPORT_UPLOAD_URL = {
    'staging': 'https://clients2.google.com/cr/staging_report',
    'prod': 'https://clients2.google.com/cr/report',
}

BOT_ID_KEY = 'botId'
CF_ID_KEY = 'clusterfuzzId'
CRASH_DUMP_PATH_MARKER = 'Output crash dump:'
PRODUCT_KEY = 'product'
PRODUCT_MAP = {
    'ANDROID': 'Chrome_Android',
    'LINUX': 'Chrome_Linux',
    'MAC': 'Chrome_Mac',
    'WINDOWS': 'Chrome',
}
PROCESSED_REPORT_FILE_KEY = 'processedReport'
MINIDUMP_FILE_KEY = 'upload_file_minidump'
VERSION_KEY = 'version'


def post_with_retries(upload_url, params, files):
  """Perform HTTP POST request to given upload url with provided params."""
  retry_limit = environment.get_value('FAIL_RETRIES')
  for _ in range(retry_limit):
    try:
      result = requests.post(upload_url, data=params, files=files)
      if result.status_code == requests.codes.ok:
        return result.text

      # No need to retry on a non-200 status code.
      logs.log_error(
          'Failed to upload request, error code %d.' % result.status_code)
      return None
    except Exception:
      # To catch cases like connection error, timeout error, etc.
      logs.log_warn('Failed to upload request, retrying.')

  logs.log_error('Could not upload request after retries.')
  return None


class FileMetadataInfo(object):
  """Handles file metadata for e.g. minidumps and processed reports."""

  def __init__(self, path=None, key=None, contents=None):
    self._path = path
    self._key = key
    self._contents = contents

  @property
  def path(self):
    return self._path

  @path.setter
  def path(self, local_path):
    self._path = local_path

  @property
  def key(self):
    return self._key

  @key.setter
  def key(self, key):
    self._key = key

  @property
  def contents(self):
    return self._contents

  @contents.setter
  def contents(self, contents):
    self._contents = contents

  def get_file_handle(self):
    """Return file handle to metadata contents. Prefer to use blobstore key if
       available, otherwise raw contents."""
    if self.key:
      contents = blobs.read_key(self.key)
    elif self.contents:
      contents = self.contents
    else:
      # No bot-independent file for which to get a file handle. Let the caller
      # handle any errors.
      return None

    metadata_file = tempfile.TemporaryFile()
    metadata_file.write(contents)
    metadata_file.seek(0)
    return metadata_file


class CrashReportInfo(object):
  """Stores the data collected from a run (via stacktrace or other) to be
     used in uploading to Chromecrash."""

  def __init__(self,
               minidump_path=None,
               minidump_contents=None,
               minidump_key=None,
               product=None,
               version=None,
               optional_params=None,
               unsymbolized_stacktrace=None,
               symbolized_stacktrace=None,
               testcase_id=None,
               bot_id=None,
               serialized_crash_stack_frames=None):
    self._minidump_info = FileMetadataInfo(
        path=minidump_path, contents=minidump_contents, key=minidump_key)
    self._minidump_path = minidump_path
    self._product = product
    self._version = version
    if optional_params is None:
      self._optional_params = {}
    else:
      self._optional_params = optional_params
    self._unsymbolized_stacktrace = unsymbolized_stacktrace
    self._symbolized_stacktrace = symbolized_stacktrace
    self._testcase_id = testcase_id
    self._bot_id = bot_id
    self._serialized_crash_stack_frames = serialized_crash_stack_frames

  @property
  def minidump_info(self):
    return self._minidump_info

  @minidump_info.setter
  def minidump_info(self, minidump_info):
    self._minidump_info = minidump_info

  @property
  def product(self):
    return self._product

  @product.setter
  def product(self, product):
    self._product = product

  @property
  def version(self):
    return self._version

  @version.setter
  def version(self, version):
    self._version = version

  @property
  def optional_params(self):
    return self._optional_params

  @optional_params.setter
  def optional_params(self, params):
    self._optional_params = params

  @property
  def unsymbolized_stacktrace(self):
    return self._unsymbolized_stacktrace

  @unsymbolized_stacktrace.setter
  def unsymbolized_stacktrace(self, stacktrace):
    self._unsymbolized_stacktrace = stacktrace

  @property
  def symbolized_stacktrace(self):
    return self._symbolized_stacktrace

  @symbolized_stacktrace.setter
  def symbolized_stacktrace(self, stacktrace):
    self._symbolized_stacktrace = stacktrace

  @property
  def testcase_id(self):
    return self._testcase_id

  @testcase_id.setter
  def testcase_id(self, testcase_id):
    self._testcase_id = testcase_id

  @property
  def bot_id(self):
    return self._bot_id

  @bot_id.setter
  def bot_id(self, bot_id):
    self._bot_id = bot_id

  @property
  def serialized_crash_stack_frames(self):
    return self._serialized_crash_stack_frames

  @serialized_crash_stack_frames.setter
  def serialized_crash_stack_frames(self, serialized_crash_stack_frames):
    self._serialized_crash_stack_frames = serialized_crash_stack_frames

  def upload(self):
    """Upload the minidump represented by self, with any other params to send
       along with the POST request."""
    if self.product is None or self.version is None:
      logs.log_error('Missing product/version info, cannot upload.')
      return None

    report_file = FileMetadataInfo(
        contents=self.serialized_crash_stack_frames).get_file_handle()
    if report_file is None:
      logs.log_warn('Missing processed report, falling back to minidump.')

    # Get minidump if there is one, but don't worry if there isn't. Just having
    # a report file is fine.
    minidump_file = self.minidump_info.get_file_handle()

    if report_file is None and minidump_file is None:
      logs.log_error('Neither processed report nor minidump, nothing to '
                     'upload.')
      return None

    # Build the upload parameters.
    params = {}
    params[PRODUCT_KEY] = self.product
    params[VERSION_KEY] = self.version
    if self.testcase_id is not None:
      params[CF_ID_KEY] = self.testcase_id
    if self.bot_id is not None:
      params[BOT_ID_KEY] = self.bot_id

    files = {}
    if report_file is not None:
      files[PROCESSED_REPORT_FILE_KEY] = report_file
    if minidump_file is not None:
      files[MINIDUMP_FILE_KEY] = minidump_file

    # Send off the report, returning the report ID.
    mode = environment.get_value('UPLOAD_MODE')
    if not mode or mode not in CRASH_REPORT_UPLOAD_URL:
      logs.log_warn(
          'Missing or unknown mode (%s); uploading to staging.' % str(mode))
      mode = 'staging'
    return post_with_retries(CRASH_REPORT_UPLOAD_URL[mode], params, files)

  def store_minidump(self):
    """Store the crash minidump in appengine and return key."""
    if not self.minidump_info.path:
      return ''

    minidump_key = ''
    logs.log('Storing minidump (%s) in blobstore.' % self.minidump_info.path)
    try:
      minidump_key = ''
      with open(self.minidump_info.path, 'rb') as file_handle:
        minidump_key = blobs.write_blob(file_handle)
    except:
      logs.log_error('Failed to store minidump.')

    if minidump_key:
      self.minidump_info = FileMetadataInfo(
          path=self.minidump_info.path, key=minidump_key)

    return minidump_key

  def to_report_metadata(self):
    """Export to ReportMetadata for batching upload."""
    return data_types.ReportMetadata(
        product=self.product,
        version=str(self.version),
        minidump_key=self.minidump_info.key,
        serialized_crash_stack_frames=self.serialized_crash_stack_frames,
        testcase_id=str(self.testcase_id),
        bot_id=str(self.bot_id))


def crash_report_info_from_metadata(report_metadata):
  """Return CrashReportInfo given ReportMetadata for uploading."""
  return CrashReportInfo(
      product=report_metadata.product,
      version=report_metadata.version,
      minidump_key=report_metadata.minidump_key,
      serialized_crash_stack_frames=(
          report_metadata.serialized_crash_stack_frames),
      testcase_id=(report_metadata.testcase_id or None),
      bot_id=(report_metadata.bot_id or None))


def parse_mime_to_crash_report_info(local_minidump_mime_path):
  """Read the (local) minidump MIME file into a CrashReportInfo object."""
  # Get the minidump name and path.
  minidump_path_match = re.match(r'(.*)\.mime', local_minidump_mime_path)
  if minidump_path_match is None:
    logs.log_error('Minidump filename in unexpected format: \'%s\'.' %
                   local_minidump_mime_path)
    return None
  minidump_path = '%s.dmp' % minidump_path_match.group(1).strip()

  # Reformat the minidump MIME to include the boundary.
  with open(local_minidump_mime_path, 'rb') as minidump_mime_file_content:
    # The boundary is the first line after the first two dashes.
    boundary = minidump_mime_file_content.readline().strip()[2:]
    minidump_mime_bytes = (
        b'Content-Type: multipart/form-data; boundary=\"%s\"\r\n--%s\r\n' %
        (boundary, boundary))
    minidump_mime_bytes += minidump_mime_file_content.read()

  minidump_mime_contents = email.message_from_bytes(minidump_mime_bytes)

  # Parse the MIME contents, extracting the parameters needed for upload.
  mime_key_values = {}
  for mime_part in minidump_mime_contents.get_payload():
    if isinstance(mime_part, str):
      mime_part = utils.decode_to_unicode(mime_part)
      logs.log_error('Unexpected str mime_part from mime path %s: %s' %
                     (local_minidump_mime_path, mime_part))
      continue
    part_descriptor = list(mime_part.values())
    key_tokens = part_descriptor[0].split('; ')
    key_match = re.match(r'name="(.*)".*', key_tokens[1])

    # Extract from the MIME part the key-value pairs used by report uploading.
    if key_match is not None:
      report_key = key_match.group(1)
      report_value = mime_part.get_payload(decode=True)
      if report_key == MINIDUMP_FILE_KEY:
        utils.write_data_to_file(report_value, minidump_path)
      else:
        # Take care of aliases.
        if report_key in ('prod', 'buildTargetId'):
          report_key = PRODUCT_KEY
        elif report_key == 'ver':
          report_key = VERSION_KEY

        # Save the key-value pair.
        mime_key_values[report_key] = report_value

  # Pull out product and version explicitly since these are required
  # for upload.
  product, version = None, None
  if PRODUCT_KEY in mime_key_values:
    product = mime_key_values.pop(PRODUCT_KEY).decode('utf-8')
  else:
    logs.log_error(
        'Could not find \'%s\' or alias in mime_key_values key.' % PRODUCT_KEY)
  if VERSION_KEY in mime_key_values:
    version = mime_key_values.pop(VERSION_KEY).decode('utf-8')
  else:
    logs.log_error(
        'Could not find \'%s\' or alias in mime_key_values key.' % VERSION_KEY)

  # If missing, return None and log keys that do exist; otherwise, construct
  # CrashReportInfo and return.
  if product is None or version is None:
    logs.log_error(
        'mime_key_values dict keys:\n%s' % str(list(mime_key_values.keys())))
    return None

  return CrashReportInfo(
      minidump_path=minidump_path,
      product=product,
      version=version,
      optional_params=mime_key_values)


def get_crash_info(output):
  """Parse crash output to get (local) minidump path and any other information
     useful for crash uploading, and store in a CrashReportInfo object."""
  crash_stacks_directory = environment.get_value('CRASH_STACKTRACES_DIR')

  output_lines = output.splitlines()
  num_lines = len(output_lines)
  is_android = environment.is_android()
  for i, line in enumerate(output_lines):
    if is_android:
      # If we are on Android, the dump extraction is more complicated.
      # The location placed in the crash-stacktrace is of the dump itself but
      # in fact only the MIME of the dump exists, and will have a different
      # extension. We need to pull the MIME and process it.
      match = re.match(CRASH_DUMP_PATH_MARKER, line)
      if not match:
        continue

      minidump_mime_filename_base = None
      for j in range(i + 1, num_lines):
        line = output_lines[j]
        match = re.match(r'(.*)\.dmp', line)
        if match:
          minidump_mime_filename_base = os.path.basename(match.group(1).strip())
          break
      if not minidump_mime_filename_base:
        logs.log_error('Minidump marker was found, but no path in stacktrace.')
        return None

      # Look for MIME. If none found, bail.
      # We might not have copied over the crash dumps yet (copying is buffered),
      # so we want to search both the original directory and the one to which
      # the minidumps should later be copied.
      device_directories_to_search = [
          constants.CRASH_DUMPS_DIR,
          os.path.dirname(line.strip())
      ]
      device_minidump_search_paths = []
      device_minidump_mime_path = None

      for device_directory in device_directories_to_search:
        device_minidump_mime_potential_paths = adb.run_shell_command(
            ['ls', '"%s"' % device_directory], root=True).splitlines()
        device_minidump_search_paths += device_minidump_mime_potential_paths

        for potential_path in device_minidump_mime_potential_paths:
          # Check that we actually found a file, and the right one (not logcat).
          if 'No such file or directory' in potential_path:
            continue

          if minidump_mime_filename_base not in potential_path:
            continue

          if '.up' in potential_path or '.dmp' in potential_path:
            device_minidump_mime_path = os.path.join(device_directory,
                                                     potential_path)
            break

        # Break if we found a path.
        if device_minidump_mime_path is not None:
          break

      # If we still didn't find a minidump path, bail.
      if device_minidump_mime_path is None:
        logs.log_error('Could not get MIME path from ls:\n%s' %
                       str(device_minidump_search_paths))
        return None

      # Pull out MIME and parse to minidump file and MIME parameters.
      minidump_mime_filename = '%s.mime' % minidump_mime_filename_base
      local_minidump_mime_path = os.path.join(crash_stacks_directory,
                                              minidump_mime_filename)
      adb.run_command([
          'pull',
          '"%s"' % device_minidump_mime_path, local_minidump_mime_path
      ])
      if not os.path.exists(local_minidump_mime_path):
        logs.log_error('Could not pull MIME from %s to %s.' %
                       (device_minidump_mime_path, local_minidump_mime_path))
        return None

      crash_info = parse_mime_to_crash_report_info(local_minidump_mime_path)
      if crash_info is None:
        return None

      crash_info.unsymbolized_stacktrace = output
      return crash_info

    # Other platforms are not currently supported.
    logs.log_error('Unable to fetch crash information for this platform.')
    return None

  # Could not find dump location, bail out. This could also happen when we don't
  # have a minidump location in stack at all, e.g. when testcase does not crash
  # during minimization.
  return None


def get_crash_info_and_stacktrace(application_command_line, crash_stacktrace,
                                  gestures):
  """Return crash minidump location and updated crash stacktrace."""
  app_name_lower = environment.get_value('APP_NAME').lower()
  retry_limit = environment.get_value('FAIL_RETRIES')
  using_android = environment.is_android()
  using_chrome = 'chrome' in app_name_lower or 'chromium' in app_name_lower
  warmup_timeout = environment.get_value('WARMUP_TIMEOUT', 90)

  # Minidump generation is only applicable on Chrome application.
  # FIXME: Support minidump generation on platforms other than Android.
  if not using_chrome or not using_android:
    return None, crash_stacktrace

  # Get the crash info from stacktrace.
  crash_info = get_crash_info(crash_stacktrace)

  # If we lost the minidump file, we need to recreate it.
  # Note that because of the way crash_info is generated now, if we have a
  # non-None crash_info, we should also have its minidump path; we insert
  # the check to safeguard against possibly constructing the crash_info in
  # other ways in the future that might potentially lose the minidump path.
  if not crash_info or not crash_info.minidump_info.path:
    for _ in range(retry_limit):
      _, _, output = (
          process_handler.run_process(
              application_command_line,
              timeout=warmup_timeout,
              gestures=gestures))

      crash_info = get_crash_info(output)
      if crash_info and crash_info.minidump_info.path:
        crash_stacktrace = utils.decode_to_unicode(output)
        break

    if not crash_info or not crash_info.minidump_info.path:
      # We could not regenerate a minidump for this crash.
      logs.log('Unable to regenerate a minidump for this crash.')

  return crash_info, crash_stacktrace


def get_symbolized_stack_bytes(crash_type, crash_address, symbolized_stack):
  """Get bytes for symbolized crash proto."""
  # FIXME: For some tests, crash_address is literally 'address'.
  crash_address = stack_parser.format_address_to_dec(crash_address)
  if crash_address is None:
    # Instead of ignoring such reports, we pass 0xDEADBEEF address.
    # Crash address is not applicable to many crash types (e.g. CHECK failures,
    # MSan and UBSan crashes).
    crash_address = 0xDEADBEEF

  # TODO(jchinlee): Add os[_info] and cpu[_info] in form crash/ expects.
  process_state = process_state_pb2.ProcessStateProto(
      crash=process_state_pb2.ProcessStateProto.Crash(
          reason=crash_type,
          address=stack_parser.unsigned_to_signed(crash_address)),
      requesting_thread=0,
  )

  try:
    for stack in symbolized_stack:
      thread = process_state.threads.add()
      thread.frames.extend([stackframe.to_proto() for stackframe in stack])

    return process_state.SerializeToString()
  except Exception as e:
    logs.log_error(
        'Failed to get proto for crash:\n'
        'Error: %s\n'
        'Type: %s\n'
        'Address: %s\n'
        'Stack:\n%s\n' % (e, crash_type, crash_address, symbolized_stack))
    return None


def save_crash_info_if_needed(testcase_id, crash_revision, job_type, crash_type,
                              crash_address, crash_frames):
  """Saves crash report for chromium project, skip otherwise."""
  if data_handler.get_project_name(job_type) != 'chromium':
    return None

  serialized_crash_stack_frames = get_symbolized_stack_bytes(
      crash_type, crash_address, crash_frames)
  if not serialized_crash_stack_frames:
    return None

  crash_info = CrashReportInfo(
      serialized_crash_stack_frames=serialized_crash_stack_frames)

  # Get product and version (required).
  platform = environment.platform()
  crash_info.product = PRODUCT_MAP[platform]
  crash_info.version = revisions.get_real_revision(
      crash_revision, job_type, display=True)

  # Update crash_info object with bot information and testcase id.
  crash_info.bot_id = environment.get_value('BOT_NAME')
  crash_info.testcase_id = int(testcase_id)

  # Store CrashInfo metadata.
  crash_report_metadata = crash_info.to_report_metadata()
  crash_report_metadata.job_type = job_type
  crash_report_metadata.crash_revision = crash_revision
  crash_report_metadata.put()

  logs.log('Created crash report entry for testcase %s.' % testcase_id)
  return crash_info



# -*- coding: utf-8 -*-
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""search_tokenizer tests."""
# pylint: disable=protected-access
import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import search_tokenizer


class ComplexTokenizeTest(unittest.TestCase):
  """Test _complex_tokenize(..)."""

  def test_empty(self):
    """Test empty string."""
    self.assertSetEqual(set(), search_tokenizer._complex_tokenize('', 3))

  def test_one_token(self):
    """Test one token."""
    self.assertSetEqual(
        set(['abcd']), search_tokenizer._complex_tokenize('abcd', 3))

  def test_multiple_tokens(self):
    """Test multiple tokens."""
    self.assertSetEqual(
        set([
            'abcd', 'abcd::edfg', 'abcd::edfghijk', 'edfg', 'edfghijk', 'hijk'
        ]), search_tokenizer._complex_tokenize('abcd::edfgHijk', 3))

  def test_multple_tokens_with_empty_tokens(self):
    """Test multiple tokens with empty tokens."""
    self.assertSetEqual(
        set([
            '::abcd', '::abcd::edfg', '::abcd::edfghijk', '::abcd::edfghijk::',
            'abcd', 'abcd::edfg', 'abcd::edfghijk', 'abcd::edfghijk::', 'edfg',
            'edfghijk', 'hijk', 'edfghijk::', 'hijk::'
        ]), search_tokenizer._complex_tokenize('::abcd::edfgHijk::', 5))

  def test_real_example(self):
    """Test real example."""
    crash_state = 'void WTF::Vector<blink::Member, 64ul'
    expected = set([
        'void',
        'void wtf',
        'void wtf::vector',
        'void wtf::vector<blink',
        'void wtf::vector<blink::member',
        'void wtf::vector<blink::member, 64ul',
        'wtf',
        'wtf::vector',
        'wtf::vector<blink',
        'wtf::vector<blink::member',
        'wtf::vector<blink::member, 64ul',
        'vector',
        'vector<blink',
        'vector<blink::member',
        'vector<blink::member, 64ul',
        'blink',
        'blink::member',
        'blink::member, 64ul',
        'member',
        'member, 64ul',
        '64ul',
    ])
    self.assertSetEqual(expected,
                        search_tokenizer._complex_tokenize(crash_state, 6))

  def test_duplicate(self):
    """Test duplicate tokens."""
    crash_state = 'a:b:a:b'
    expected = set(['a', 'b', 'a:b', 'a:b:a', 'a:b:a:b', 'b:a', 'b:a:b'])
    self.assertSetEqual(expected,
                        search_tokenizer._complex_tokenize(crash_state, 4))

  def test_exceed_limit(self):
    """Test exceeding limit."""
    crash_state = 'a:b:c'
    expected = set(['a', 'b', 'c', 'a:b', 'b:c'])
    self.assertSetEqual(expected,
                        search_tokenizer._complex_tokenize(crash_state, 2))


class TokenizeBugInformationTest(unittest.TestCase):
  """Test tokenize_bug_information(..)."""

  def test_none(self):
    """Test none."""
    testcase = data_types.Testcase()
    testcase.bug_information = None
    testcase.group_bug_information = None
    self.assertListEqual([],
                         search_tokenizer.tokenize_bug_information(testcase))

  def test_empty(self):
    """Test empty."""
    testcase = data_types.Testcase()
    testcase.bug_information = ''
    testcase.group_bug_information = 0
    self.assertListEqual([],
                         search_tokenizer.tokenize_bug_information(testcase))

  def test_tokenize(self):
    """Test tokenize."""
    testcase = data_types.Testcase()
    testcase.bug_information = '123'
    testcase.group_bug_information = 456
    self.assertListEqual(['123', '456'],
                         search_tokenizer.tokenize_bug_information(testcase))


class TokenizeImpactVersionTest(unittest.TestCase):
  """Test tokenize_impact(..)."""

  def test_empty(self):
    """Test empty."""
    self.assertEqual([], search_tokenizer.tokenize_impact_version(''))
    self.assertEqual([], search_tokenizer.tokenize_impact_version(None))

  def test_version(self):
    """Test tokenising version."""
    self.assertEqual(['52'], search_tokenizer.tokenize_impact_version('52'))
    self.assertSetEqual(
        set(['52', '52.1', '52.1.2', '52.1.2.3']),
        set(search_tokenizer.tokenize_impact_version('52.1.2.3')))


class TokenizeTest(unittest.TestCase):
  """Test tokenize(..)."""

  def test_empty(self):
    """Test empty string."""
    self.assertSetEqual(set(), search_tokenizer.tokenize(''))

  def test_non_string(self):
    """Test non string."""
    self.assertSetEqual(set(['123']), search_tokenizer.tokenize(123))
    self.assertSetEqual(set(['true']), search_tokenizer.tokenize(True))
    self.assertSetEqual(set([]), search_tokenizer.tokenize(None))

  def test_non_ascii(self):
    s = 'IsString ¿ÓÞÎ¤ utf'
    self.assertSetEqual(
        set([
            'is', 'string', 'utf', 'isstring', 'isstring utf', 'string utf',
            s.lower()
        ]), search_tokenizer.tokenize(s))

  def test_real_example(self):
    """Test real example."""
    crash_states = '\n'.join([
        'track 1 fast;',
        'android.media.MediaCodec.native_setup',
    ])
    expected = set([
        'track',
        '1',
        'fast',
        'android',
        'media',
        'codec',
        'native',
        'setup',
        'track 1',
        'track 1 fast',
        'track 1 fast;',
        '1 fast',
        '1 fast;',
        'fast;',
        'android.media',
        'android.media.media',
        'android.media.mediacodec',
        'android.media.mediacodec.native',
        'android.media.mediacodec.native_setup',
        'media.media',
        'media.mediacodec',
        'media.mediacodec.native',
        'media.mediacodec.native_setup',
        'mediacodec',
        'mediacodec.native',
        'mediacodec.native_setup',
        'codec.native',
        'codec.native_setup',
        'native_setup',
    ])
    self.assertSetEqual(expected, search_tokenizer.tokenize(crash_states))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Analyze task for handling user uploads."""

import datetime

import six

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.bot.tasks import task_creation
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.chrome import crash_uploader
from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.crash_analysis import severity_analyzer
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.fuzzing import leak_blacklist
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment


def _add_default_issue_metadata(testcase):
  """Adds the default issue metadata (e.g. components, labels) to testcase."""
  default_metadata = engine_common.get_all_issue_metadata_for_testcase(testcase)
  if not default_metadata:
    return

  testcase_metadata = testcase.get_metadata()
  for key, default_value in six.iteritems(default_metadata):
    # Add the default issue metadata first. This gives preference to uploader
    # specified issue metadata.
    new_value_list = utils.parse_delimited(
        default_value, delimiter=',', strip=True, remove_empty=True)

    # Append uploader specified testcase metadata value to end (for preference).
    uploader_value = testcase_metadata.get(key, '')
    uploader_value_list = utils.parse_delimited(
        uploader_value, delimiter=',', strip=True, remove_empty=True)
    for value in uploader_value_list:
      if value not in new_value_list:
        new_value_list.append(value)

    new_value = ','.join(new_value_list)
    if new_value == uploader_value:
      continue

    logs.log('Updating issue metadata for {} from {} to {}.'.format(
        key, uploader_value, new_value))
    testcase.set_metadata(key, new_value)


def setup_build(testcase):
  """Set up a custom or regular build based on revision. For regular builds,
  if a provided revision is not found, set up a build with the
  closest revision <= provided revision."""
  revision = testcase.crash_revision

  if revision and not build_manager.is_custom_binary():
    build_bucket_path = build_manager.get_primary_bucket_path()
    revision_list = build_manager.get_revisions_list(
        build_bucket_path, testcase=testcase)
    if not revision_list:
      logs.log_error('Failed to fetch revision list.')
      return

    revision_index = revisions.find_min_revision_index(revision_list, revision)
    if revision_index is None:
      raise errors.BuildNotFoundError(revision, testcase.job_type)
    revision = revision_list[revision_index]

  build_manager.setup_build(revision)


def execute_task(testcase_id, job_type):
  """Run analyze task."""
  # Reset redzones.
  environment.reset_current_memory_tool_options(redzone_size=128)

  # Unset window location size and position properties so as to use default.
  environment.set_value('WINDOW_ARG', '')

  # Locate the testcase associated with the id.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  if not testcase:
    return

  data_handler.update_testcase_comment(testcase, data_types.TaskState.STARTED)

  metadata = data_types.TestcaseUploadMetadata.query(
      data_types.TestcaseUploadMetadata.testcase_id == int(testcase_id)).get()
  if not metadata:
    logs.log_error(
        'Testcase %s has no associated upload metadata.' % testcase_id)
    testcase.key.delete()
    return

  is_lsan_enabled = environment.get_value('LSAN')
  if is_lsan_enabled:
    # Creates empty local blacklist so all leaks will be visible to uploader.
    leak_blacklist.create_empty_local_blacklist()

  # Store the bot name and timestamp in upload metadata.
  bot_name = environment.get_value('BOT_NAME')
  metadata.bot_name = bot_name
  metadata.timestamp = datetime.datetime.utcnow()
  metadata.put()

  # Adjust the test timeout, if user has provided one.
  if metadata.timeout:
    environment.set_value('TEST_TIMEOUT', metadata.timeout)

  # Adjust the number of retries, if user has provided one.
  if metadata.retries is not None:
    environment.set_value('CRASH_RETRIES', metadata.retries)

  # Set up testcase and get absolute testcase path.
  file_list, _, testcase_file_path = setup.setup_testcase(testcase, job_type)
  if not file_list:
    return

  # Set up build.
  setup_build(testcase)

  # Check if we have an application path. If not, our build failed
  # to setup correctly.
  if not build_manager.check_app_path():
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         'Build setup failed')

    if data_handler.is_first_retry_for_task(testcase):
      build_fail_wait = environment.get_value('FAIL_WAIT')
      tasks.add_task(
          'analyze', testcase_id, job_type, wait_time=build_fail_wait)
    else:
      data_handler.close_invalid_uploaded_testcase(testcase, metadata,
                                                   'Build setup failed')
    return

  # Update initial testcase information.
  testcase.absolute_path = testcase_file_path
  testcase.job_type = job_type
  testcase.binary_flag = utils.is_binary_file(testcase_file_path)
  testcase.queue = tasks.default_queue()
  testcase.crash_state = ''

  # Set initial testcase metadata fields (e.g. build url, etc).
  data_handler.set_initial_testcase_metadata(testcase)

  # Update minimized arguments and use ones provided during user upload.
  if not testcase.minimized_arguments:
    minimized_arguments = environment.get_value('APP_ARGS') or ''
    additional_command_line_flags = testcase.get_metadata(
        'uploaded_additional_args')
    if additional_command_line_flags:
      minimized_arguments += ' %s' % additional_command_line_flags
    environment.set_value('APP_ARGS', minimized_arguments)
    testcase.minimized_arguments = minimized_arguments

  # Update other fields not set at upload time.
  testcase.crash_revision = environment.get_value('APP_REVISION')
  data_handler.set_initial_testcase_metadata(testcase)
  testcase.put()

  # Initialize some variables.
  gestures = testcase.gestures
  http_flag = testcase.http_flag
  test_timeout = environment.get_value('TEST_TIMEOUT')

  # Get the crash output.
  result = testcase_manager.test_for_crash_with_retries(
      testcase,
      testcase_file_path,
      test_timeout,
      http_flag=http_flag,
      compare_crash=False)

  # If we don't get a crash, try enabling http to see if we can get a crash.
  # Skip engine fuzzer jobs (e.g. libFuzzer, AFL) for which http testcase paths
  # are not applicable.
  if (not result.is_crash() and not http_flag and
      not environment.is_engine_fuzzer_job()):
    result_with_http = testcase_manager.test_for_crash_with_retries(
        testcase,
        testcase_file_path,
        test_timeout,
        http_flag=True,
        compare_crash=False)
    if result_with_http.is_crash():
      logs.log('Testcase needs http flag for crash.')
      http_flag = True
      result = result_with_http

  # Refresh our object.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  if not testcase:
    return

  # Set application command line with the correct http flag.
  application_command_line = (
      testcase_manager.get_command_line_for_application(
          testcase_file_path, needs_http=http_flag))

  # Get the crash data.
  crashed = result.is_crash()
  crash_time = result.get_crash_time()
  state = result.get_symbolized_data()
  unsymbolized_crash_stacktrace = result.get_stacktrace(symbolized=False)

  # Get crash info object with minidump info. Also, re-generate unsymbolized
  # stacktrace if needed.
  crash_info, _ = (
      crash_uploader.get_crash_info_and_stacktrace(
          application_command_line, state.crash_stacktrace, gestures))
  if crash_info:
    testcase.minidump_keys = crash_info.store_minidump()

  if not crashed:
    # Could not reproduce the crash.
    log_message = (
        'Testcase didn\'t crash in %d seconds (with retries)' % test_timeout)
    data_handler.update_testcase_comment(
        testcase, data_types.TaskState.FINISHED, log_message)

    # In the general case, we will not attempt to symbolize if we do not detect
    # a crash. For user uploads, we should symbolize anyway to provide more
    # information about what might be happening.
    crash_stacktrace_output = utils.get_crash_stacktrace_output(
        application_command_line, state.crash_stacktrace,
        unsymbolized_crash_stacktrace)
    testcase.crash_stacktrace = data_handler.filter_stacktrace(
        crash_stacktrace_output)

    # For an unreproducible testcase, retry once on another bot to confirm
    # our results and in case this bot is in a bad state which we didn't catch
    # through our usual means.
    if data_handler.is_first_retry_for_task(testcase):
      testcase.status = 'Unreproducible, retrying'
      testcase.put()

      tasks.add_task('analyze', testcase_id, job_type)
      return

    data_handler.close_invalid_uploaded_testcase(testcase, metadata,
                                                 'Unreproducible')

    # A non-reproducing testcase might still impact production branches.
    # Add the impact task to get that information.
    task_creation.create_impact_task_if_needed(testcase)
    return

  # Update testcase crash parameters.
  testcase.http_flag = http_flag
  testcase.crash_type = state.crash_type
  testcase.crash_address = state.crash_address
  testcase.crash_state = state.crash_state
  crash_stacktrace_output = utils.get_crash_stacktrace_output(
      application_command_line, state.crash_stacktrace,
      unsymbolized_crash_stacktrace)
  testcase.crash_stacktrace = data_handler.filter_stacktrace(
      crash_stacktrace_output)

  # Try to guess if the bug is security or not.
  security_flag = crash_analyzer.is_security_issue(
      state.crash_stacktrace, state.crash_type, state.crash_address)
  testcase.security_flag = security_flag

  # If it is, guess the severity.
  if security_flag:
    testcase.security_severity = severity_analyzer.get_security_severity(
        state.crash_type, state.crash_stacktrace, job_type, bool(gestures))

  log_message = ('Testcase crashed in %d seconds (r%d)' %
                 (crash_time, testcase.crash_revision))
  data_handler.update_testcase_comment(testcase, data_types.TaskState.FINISHED,
                                       log_message)

  # See if we have to ignore this crash.
  if crash_analyzer.ignore_stacktrace(state.crash_stacktrace):
    data_handler.close_invalid_uploaded_testcase(testcase, metadata,
                                                 'Irrelavant')
    return

  # Test for reproducibility.
  one_time_crasher_flag = not testcase_manager.test_for_reproducibility(
      testcase.fuzzer_name, testcase.actual_fuzzer_name(), testcase_file_path,
      state.crash_state, security_flag, test_timeout, http_flag, gestures)
  testcase.one_time_crasher_flag = one_time_crasher_flag

  # Check to see if this is a duplicate.
  data_handler.check_uploaded_testcase_duplicate(testcase, metadata)

  # Set testcase and metadata status if not set already.
  if testcase.status == 'Duplicate':
    # For testcase uploaded by bots (with quiet flag), don't create additional
    # tasks.
    if metadata.quiet_flag:
      data_handler.close_invalid_uploaded_testcase(testcase, metadata,
                                                   'Duplicate')
      return
  else:
    # New testcase.
    testcase.status = 'Processed'
    metadata.status = 'Confirmed'

    # Reset the timestamp as well, to respect
    # data_types.MIN_ELAPSED_TIME_SINCE_REPORT. Otherwise it may get filed by
    # triage task prematurely without the grouper having a chance to run on this
    # testcase.
    testcase.timestamp = utils.utcnow()

    # Add new leaks to global blacklist to avoid detecting duplicates.
    # Only add if testcase has a direct leak crash and if it's reproducible.
    if is_lsan_enabled:
      leak_blacklist.add_crash_to_global_blacklist_if_needed(testcase)

  # Update the testcase values.
  testcase.put()

  # Update the upload metadata.
  metadata.security_flag = security_flag
  metadata.put()

  _add_default_issue_metadata(testcase)

  # Create tasks to
  # 1. Minimize testcase (minimize).
  # 2. Find regression range (regression).
  # 3. Find testcase impact on production branches (impact).
  # 4. Check whether testcase is fixed (progression).
  # 5. Get second stacktrace from another job in case of
  #    one-time crashes (stack).
  task_creation.create_tasks(testcase)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Linux Kernel Library kernel utils functions."""
import os

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.platforms.android import symbols_downloader
from clusterfuzz._internal.system import environment

from . import constants


def _should_download_symbols():
  """Return True if we should continue to download symbols."""
  # For local testing or when running the reproduce tool locally, we do not
  # have access to the cloud storage bucket with the symbols. In this case,
  # just bail out.
  return (not environment.get_value('LOCAL_DEVELOPMENT') and
          not environment.get_value('REPRODUCE_TOOL'))


def get_kernel_prefix_and_full_hash(build_id):
  """Download repo.prop and return the full hash and prefix."""
  android_kernel_repo_data = _get_repo_prop_data(build_id,
                                                 constants.LKL_BUILD_TARGET)
  if android_kernel_repo_data:
    for line in android_kernel_repo_data.splitlines():
      if line.startswith(constants.LKL_REPO_KERNEL_PREFIX):
        # line is of form: prefix u'hash'
        return (constants.LKL_REPO_KERNEL_PREFIX, line.split(' ',
                                                             1)[1].strip('u\''))

  return None, None


def _get_repo_prop_data(build_id, fuzz_target):
  """Downloads repo.prop and returuns the data based on build_id and target."""
  symbols_directory = os.path.join(
      environment.get_value('SYMBOLS_DIR'), fuzz_target)
  repro_filename = symbols_downloader.get_repo_prop_archive_filename(
      build_id, fuzz_target)

  # Grab repo.prop, it is not on the device nor in the build_dir.
  _download_kernel_repo_prop_if_needed(symbols_directory, build_id, fuzz_target)
  local_repo_path = utils.find_binary_path(symbols_directory, repro_filename)
  if local_repo_path and os.path.exists(local_repo_path):
    return utils.read_data_from_file(local_repo_path, eval_data=False).decode()

  return None


def _download_kernel_repo_prop_if_needed(symbols_directory, build_id,
                                         fuzz_target):
  """Downloads the repo.prop for an LKL fuzzer"""
  if not _should_download_symbols():
    return

  symbols_downloader.download_repo_prop_if_needed(
      symbols_directory, build_id, fuzz_target, [fuzz_target], 'lkl_fuzzer')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""libFuzzer fuzzer."""
from clusterfuzz._internal.bot.fuzzers import builtin
from clusterfuzz._internal.bot.fuzzers import options
from clusterfuzz._internal.bot.fuzzers.libFuzzer import constants


def get_grammar(fuzzer_path):
  """Get grammar for a given fuzz target. Return none if there isn't one."""
  fuzzer_options = options.get_fuzz_target_options(fuzzer_path)
  if fuzzer_options:
    grammar = fuzzer_options.get_grammar_options()
    if grammar:
      return grammar.get('grammar')

  return None


def get_arguments(fuzzer_path):
  """Get arguments for a given fuzz target."""
  arguments = []
  rss_limit_mb = None
  timeout = None

  fuzzer_options = options.get_fuzz_target_options(fuzzer_path)

  if fuzzer_options:
    libfuzzer_arguments = fuzzer_options.get_engine_arguments('libfuzzer')
    if libfuzzer_arguments:
      arguments.extend(libfuzzer_arguments.list())
      rss_limit_mb = libfuzzer_arguments.get('rss_limit_mb', constructor=int)
      timeout = libfuzzer_arguments.get('timeout', constructor=int)

  if not timeout:
    arguments.append(
        '%s%d' % (constants.TIMEOUT_FLAG, constants.DEFAULT_TIMEOUT_LIMIT))
  else:
    # Custom timeout value shouldn't be greater than the default timeout
    # limit.
    # TODO(mmoroz): Eventually, support timeout values greater than the
    # default.
    if timeout > constants.DEFAULT_TIMEOUT_LIMIT:
      arguments.remove('%s%d' % (constants.TIMEOUT_FLAG, timeout))
      arguments.append(
          '%s%d' % (constants.TIMEOUT_FLAG, constants.DEFAULT_TIMEOUT_LIMIT))

  if not rss_limit_mb:
    arguments.append(
        '%s%d' % (constants.RSS_LIMIT_FLAG, constants.DEFAULT_RSS_LIMIT_MB))
  else:
    # Custom rss_limit_mb value shouldn't be greater than the default value.
    if rss_limit_mb > constants.DEFAULT_RSS_LIMIT_MB:
      arguments.remove('%s%d' % (constants.RSS_LIMIT_FLAG, rss_limit_mb))
      arguments.append(
          '%s%d' % (constants.RSS_LIMIT_FLAG, constants.DEFAULT_RSS_LIMIT_MB))

  return arguments


class LibFuzzer(builtin.EngineFuzzer):
  """Builtin libFuzzer fuzzer."""

  def generate_arguments(self, fuzzer_path):
    """Generate arguments for fuzzer using .options file or default values."""
    return ' '.join(get_arguments(fuzzer_path))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for build_setup."""

import os
import unittest

from clusterfuzz._internal.bot.untrusted_runner import build_setup
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


def _failed_setup(*_):
  return False


def _mock_regular_build_setup(*_):
  os.environ['APP_PATH'] = '/release/bin/app'
  os.environ['APP_PATH_DEBUG'] = ''
  os.environ['APP_DIR'] = '/release/bin'
  os.environ['BUILD_DIR'] = '/release'
  os.environ['BUILD_URL'] = 'https://build/url.zip'
  return True


class BuildSetupTest(unittest.TestCase):
  """Tests for build setup (untrusted side)."""

  def setUp(self):
    test_helpers.patch(self, [
        ('regular_build_setup',
         'clusterfuzz._internal.build_management.build_manager.RegularBuild.setup'
        ),
    ])

    test_helpers.patch_environ(self)

  def test_setup_regular_build(self):
    """Test setup_regular_build."""
    request = untrusted_runner_pb2.SetupRegularBuildRequest(
        base_build_dir='/base',
        revision=1337,
        build_url='https://build/url.zip',
        target_weights={
            'bad_target': 0.1,
            'normal_target': 1.0
        })

    self.mock.regular_build_setup.side_effect = _mock_regular_build_setup
    response = build_setup.setup_regular_build(request)
    self.assertTrue(response.result)
    self.assertEqual(response.app_path, '/release/bin/app')
    self.assertEqual(response.app_path_debug, '')
    self.assertEqual(response.app_dir, '/release/bin')
    self.assertEqual(response.build_dir, '/release')
    self.assertEqual(response.build_url, 'https://build/url.zip')

    self.mock.regular_build_setup.side_effect = _failed_setup
    response = build_setup.setup_regular_build(request)
    self.assertFalse(response.result)
    self.assertFalse(response.HasField('app_path'))
    self.assertFalse(response.HasField('app_path_debug'))
    self.assertFalse(response.HasField('app_dir'))
    self.assertFalse(response.HasField('build_dir'))
    self.assertFalse(response.HasField('build_url'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Coverage Report viewer."""

import datetime
import re

from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import fuzzer_stats
from handlers import base_handler
from libs import handler
from libs import helpers

VALID_DATE_REGEX = re.compile(r'^([0-9\-]+|latest)$')


def _get_project_report_url(job, date):
  """Return url for the report requested."""
  project = data_handler.get_project_name(job)
  if not project:
    return None

  if date == 'latest':
    date = None
  else:
    try:
      date = datetime.datetime.strptime(date, '%Y-%m-%d').date()
    except:
      raise helpers.EarlyExitException('Invalid date.', 400)

  info = fuzzer_stats.get_coverage_info(project, date)
  if not info:
    return None

  return info.html_report_url


def get_report_url(report_type, argument, date):
  """Get report url for a redirect from the coverage report handler."""
  # It's very easy to add support for per fuzzer reports, but we don't need it.
  if report_type != 'job':
    raise helpers.EarlyExitException('Invalid report type.', 400)

  job = argument
  if not job:
    raise helpers.EarlyExitException('Job name cannot be empty.', 400)

  if not data_types.Job.VALID_NAME_REGEX.match(job):
    raise helpers.EarlyExitException('Invalid job name.', 400)

  if not date or not VALID_DATE_REGEX.match(date):
    raise helpers.EarlyExitException('Invalid date.', 400)

  return _get_project_report_url(job, date)


class Handler(base_handler.Handler):
  """Coverage Report Handler."""

  # pylint: disable=unused-argument
  @handler.get(handler.HTML)
  def get(self, report_type=None, argument=None, date=None, extra=None):
    """Handle a get request."""
    report_url = get_report_url(report_type, argument, date)
    if report_url:
      return self.redirect(report_url)
    raise helpers.EarlyExitException('Failed to get coverage report.', 400)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stack parsing module."""
import os
import re
import string
import subprocess

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms.linux.lkl import constants as lkl_constants
from clusterfuzz._internal.system import environment

from .constants import *


class CrashInfo:
  """Parsed crash information."""

  def __init__(self):
    self.crash_type = ''
    self.crash_address = ''
    self.crash_state = ''
    self.crash_stacktrace = ''
    self.frame_count = 0
    self.process_name = 'NULL'
    self.process_died = False

    # Following fields are for internal use only and subject to change. Do not
    # rely on these.
    self.frames = []
    self.last_frame_id = -1
    self.raw_frames = []

    # Additional tracking for Java bugs.
    self.found_java_exception = False

    # Additional tracking for bad casts.
    self.found_bad_cast_crash_end_marker = False

    # Additional tracking for check failures.
    self.check_failure_source_file = ''

    # Additional tracking for fatal errors.
    self.fatal_error_occurred = False

    # Additional tracking for lkl bugs.
    self.lkl_kernel_build_id = None

    self.is_kasan = False
    self.is_lkl = False
    self.is_golang = False
    self.is_python = False
    self.found_python_crash = False
    self.found_golang_crash = False
    self.found_android_kernel_crash = False


def _filter_stack_frame(stack_frame):
  """Filter stack frame."""
  # Filter out anonymous namespaces.
  anonymous_namespaces = [
      'non-virtual thunk to ',
      '(anonymous namespace)::',
      '`anonymous namespace\'::',
  ]
  for ns in anonymous_namespaces:
    stack_frame = stack_frame.replace(ns, '')

  # Rsplit around '!'.
  stack_frame = stack_frame.split('!')[-1]

  # Lsplit around '(', '['.
  m = re.match(r'(.*?)[\(\[].*', stack_frame)
  if m and m.group(1):
    return m.group(1).strip()

  # Lsplit around ' '.
  stack_frame = stack_frame.strip().split(' ')[0]

  return stack_frame


class StackParser:
  """Stack parser."""

  def __init__(self,
               symbolized=True,
               detect_ooms_and_hangs=True,
               detect_v8_runtime_errors=False,
               custom_stack_frame_ignore_regexes=None,
               fuzz_target=None,
               include_ubsan=True):

    if not custom_stack_frame_ignore_regexes:
      custom_stack_frame_ignore_regexes = []

    self.stack_frame_ignore_regex = re.compile(
        r'(%s)' % '|'.join(STACK_FRAME_IGNORE_REGEXES +
                           custom_stack_frame_ignore_regexes))

    self.stack_frame_ignore_regex_if_symbolized = re.compile(
        r'(%s)' % '|'.join(STACK_FRAME_IGNORE_REGEXES_IF_SYMBOLIZED))

    self.detect_ooms_and_hangs = detect_ooms_and_hangs
    self.detect_v8_runtime_errors = detect_v8_runtime_errors
    self.symbolized = symbolized
    self.fuzz_target = fuzz_target
    self.include_ubsan = include_ubsan

  def ignore_stack_frame(self, stack_frame):
    """Return true if stack frame should not used in determining the
    crash state."""
    # No data, should ignore.
    if not stack_frame:
      return True

    # Too short of a stack frame, nothing to do.
    if len(stack_frame) < 3:
      return True

    # Normalize path separator in stack frame, this allows to ignore strings
    # properly cross-platform.
    normalized_stack_frame = stack_frame.replace('\\', '/')

    # Check if the stack frame matches one of the ignore list regexes.
    if self.stack_frame_ignore_regex.match(normalized_stack_frame):
      return True

    if self.symbolized and self.stack_frame_ignore_regex_if_symbolized.match(
        normalized_stack_frame):
      return True

    return False

  def update_state_on_match(self,
                            compiled_regex,
                            line,
                            state,
                            new_type=None,
                            new_state=None,
                            new_frame_count=None,
                            new_address=None,
                            address_from_group=None,
                            type_from_group=None,
                            state_from_group=None,
                            address_filter=lambda s: s,
                            type_filter=lambda s: s,
                            reset=False):
    """Update the specified parts of the state if we have a match."""

    match = compiled_regex.match(line)
    if not match:
      return None

    if reset:
      state.crash_address = ''
      state.crash_state = ''
      state.frame_count = 0

    # Direct updates.
    if new_type is not None:
      state.crash_type = new_type

    if new_state is not None:
      state.crash_state = new_state

    if new_frame_count is not None:
      state.frame_count = new_frame_count

    if new_address is not None:
      state.crash_address = new_address

    # Updates from match groups.
    if type_from_group is not None:
      state.crash_type = type_filter(match.group(type_from_group)).strip()

    if address_from_group is not None:
      state.crash_address = address_filter(
          match.group(address_from_group)).strip()

    if state_from_group is not None:
      state.crash_state = match.group(state_from_group)

    return match

  def add_frame_on_match(self,
                         compiled_regex,
                         line,
                         state,
                         group=0,
                         frame_filter=_filter_stack_frame,
                         demangle=False,
                         can_ignore=True,
                         frame_spec=None,
                         frame_override_func=None):
    """Add a frame to the crash state if we have a match on this line."""
    match = compiled_regex.match(line)
    if not match:
      return None

    frame = match.group(group).strip()

    # Strip out unneeded structure. Remove this hack after modularizing tools.
    for regex in STRIP_STRUCTURE_REGEXES:
      structure_match = regex.match(frame)
      if structure_match:
        frame = structure_match.group(1)
        break

    if demangle and environment.is_posix():
      pipe = subprocess.Popen(
          ['c++filt', '-n', frame],
          stdin=subprocess.PIPE,
          stdout=subprocess.PIPE)
      frame, _ = pipe.communicate()
      frame = frame.decode('utf-8')

    # Try to parse the frame with the various stackframes.
    frame_struct = None
    if frame_spec is not None:
      frame_struct = frame_spec.parse_stack_frame(match)

      # Account for case when we have no symbols and hence don't want to strip
      # anonymous namespaces in the crash state.
      if (frame_struct and frame_struct.module_offset and
          not frame_struct.function_name):
        frame_filter = lambda s: s

      # Update stacktrace frames list with frame struct.
      new_thread = state.last_frame_id < 0
      try:
        # We have a 'frame_id' group, so pull the ID and check if we need to
        # start a new thread. Also update last_frame_id accordingly.
        frame_id = int(match.group('frame_id').strip())
        if frame_id < state.last_frame_id:
          new_thread = True
        state.last_frame_id = frame_id
      except IndexError:
        # If there is not 'frame_id' group, just stick everything into one
        # thread.
        state.last_frame_id += 1

      if new_thread:
        state.frames.append([])
      state.frames[-1].append(frame_struct)

    if frame_override_func:
      frame = frame_override_func(frame, frame_struct)

    # If we are ignoring a frame, we still have a match. Don't add it to the
    # state, but notify the caller that we found something.
    if can_ignore and self.ignore_stack_frame(frame):
      return match

    # Filter the frame and add to a list.
    filtered_frame = frame_filter(frame)
    state.raw_frames.append(filtered_frame)

    # Update the crash state only if we need more frames.
    if state.frame_count < MAX_CRASH_STATE_FRAMES:
      state.crash_state += filtered_frame + '\n'
      state.frame_count += 1

    return match

  def update_state_on_check_failure(self, state, line, regex, crash_type):
    """Update the state if the crash is a CHECK failure."""
    check_match = self.update_state_on_match(
        regex, line, state, new_type=crash_type, reset=True, new_frame_count=1)
    if check_match:
      failure_string = fix_check_failure_string(check_match.group(2))
      source_file = fix_filename_string(check_match.group(1))
      state.crash_state = '%s in %s\n' % (failure_string, source_file)

  def match_assert(self, line, state, regex, group=1):
    """Match an assert."""
    assert_match = self.update_state_on_match(
        regex, line, state, new_type='ASSERT', new_frame_count=1)
    if assert_match and assert_match.group(group):
      # For asserts, we want to actually use the match as the crash state.
      state.crash_state = assert_match.group(group) + '\n'

  def filter_crash_parameters(self, state):
    """Normalize crash parameters into generic format regardless of the tool
    used."""
    # Filter crash state represented in |state|.
    # Remove non-printable chars from crash state.
    state.crash_state = ''.join(
        s for s in state.crash_state if s in string.printable)

    # Shorten JNI messages.
    if JNI_ERROR_STRING in state.crash_state:
      state.crash_state = state.crash_state.replace(JNI_ERROR_STRING, 'JNI:')

    if self.symbolized:
      # Normalize addresses and numbers in crash_state.
      # Skip normalization for V8 correctness failures, which use the crash
      # state to store metadata containing numbers.
      if state.crash_type not in ['V8 correctness failure']:
        state.crash_state = filter_addresses_and_numbers(state.crash_state)

      # Truncate each line in the crash state to avoid excessive length.
      original_crash_state = state.crash_state
      state.crash_state = ''
      for line in original_crash_state.splitlines():
        # Exclude bad-cast line for bad cast testcases.
        # FIXME: Find a way to make bad-cast lines shorter and then remove this.
        if line.startswith('Bad-cast'):
          state.crash_state += line + '\n'
        else:
          state.crash_state += line[:LINE_LENGTH_CAP] + '\n'

    # Don't return an empty crash state if we have a crash type. Either set
    # to NULL or use the crashing process name if available.
    if state.crash_type and not state.crash_state.strip():
      state.crash_state = state.process_name

    # For timeout, OOMs, const-input-overwrites in fuzz targets, force use of
    # fuzz target name since stack itself is not usable for deduplication.
    if self.fuzz_target and state.crash_type in [
        'Out-of-memory', 'Timeout', 'Overwrites-const-input'
    ]:
      state.crash_state = self.fuzz_target

    # Add a trailing \n if it does not exist in crash state.
    if (state.crash_state and state.crash_state != 'NULL' and
        state.crash_state[-1] != '\n'):
      state.crash_state += '\n'

    # Normalize access size parameter if greater than 16 bytes.
    m = re.match('([^0-9]+)([0-9]+)', state.crash_type, re.DOTALL)
    if m:
      num = int(m.group(2))
      if num > 16:
        num = '{*}'

      state.crash_type = (
          state.crash_type[:len(m.group(1))] + str(num) +
          state.crash_type[m.end():])

    # On some platforms crash address is unnecessarily long. We can truncate it.
    if (state.crash_address.startswith('0x0000') and
        len(state.crash_address) == 18):
      state.crash_address = '0x%s' % state.crash_address[len('0x0000'):]

    # Report null dereferences as such.
    if state.crash_address and state.crash_type.startswith('UNKNOWN'):
      int_crash_address = crash_analyzer.address_to_integer(state.crash_address)
      if crash_analyzer.is_null_dereference(int_crash_address):
        state.crash_type = state.crash_type.replace('UNKNOWN',
                                                    'Null-dereference')

    return state

  def remove_lkl_kernel_times_and_set_params(self, state, crash_data):
    """Filter kernel time from crash_data if LKL."""
    result = ''
    for line in crash_data.splitlines():
      clean_line = ANDROID_KERNEL_TIME_REGEX.sub('', line)
      result += clean_line + '\n'
      if not state.lkl_kernel_build_id:
        match = lkl_constants.LINUX_VERSION_REGEX_LKL.match(clean_line)
        if match:
          state.lkl_kernel_build_id = match.group(2)
    return result

  def parse(self, stacktrace: str) -> CrashInfo:
    """Parse a stacktrace."""
    state = CrashInfo()
    state.crash_stacktrace = stacktrace
    state.is_kasan = 'KASAN' in stacktrace
    state.is_lkl = lkl_constants.LINUX_KERNEL_MODULE_STACK_TRACE in stacktrace
    state.is_golang = '.go:' in stacktrace
    state.is_python = '.py", line' in stacktrace

    # For Android LKL (and potentially kernel output), the KASAN crash may start
    # with the time since boot.  We need to remove this so that our regexes
    # match.
    # We also want to set the kernel build id here since we are already walking
    # line by line.
    if state.is_lkl:
      stacktrace = self.remove_lkl_kernel_times_and_set_params(
          state, stacktrace)

    split_crash_stacktrace = stacktrace.splitlines()

    if state.is_python:
      split_crash_stacktrace = reverse_python_stacktrace(split_crash_stacktrace)

    for line in split_crash_stacktrace:
      if should_ignore_line_for_crash_processing(line, state):
        continue

      # Bail out from crash paramater parsing if we detect this is a
      # out-of-memory signature.
      if not self.detect_ooms_and_hangs and OUT_OF_MEMORY_REGEX.match(line):
        return CrashInfo()

      # Ignore aborts, breakpoints, ills and traps for asserts, check and
      # dcheck failures. These are intended, retain their original state.
      if (SAN_ABRT_REGEX.match(line) or SAN_BREAKPOINT_REGEX.match(line) or
          SAN_ILL_REGEX.match(line) or SAN_TRAP_REGEX.match(line)):
        if state.crash_type in IGNORE_CRASH_TYPES_FOR_ABRT_BREAKPOINT_AND_ILLS:
          continue

      # Assertions always come first, before the actual crash stacktrace.
      # However if we already have a kernel crash, we don't want to
      # replace it with the ASSERT.
      if not state.crash_type.startswith('Kernel failure'):
        self.match_assert(line, state, ASSERT_REGEX)
        self.match_assert(line, state, ASSERT_REGEX_GOOGLE, group=2)
        self.match_assert(line, state, ASSERT_REGEX_GLIBC)
        self.match_assert(line, state, RUST_ASSERT_REGEX)

      # ASSERT_NOT_REACHED prints a single line error then triggers a crash. We
      # set the crash state here, but look for the stack after a crash on an
      # unknown address.
      self.update_state_on_match(
          ASSERT_NOT_REACHED_REGEX,
          line,
          state,
          new_type='ASSERT_NOT_REACHED',
          reset=True)

      # Platform specific: Linux gdb crash type format.
      self.update_state_on_match(
          LINUX_GDB_CRASH_TYPE_REGEX,
          line,
          state,
          type_from_group=1,
          type_filter=lambda s: s.upper())

      # Platform specific: Linux gdb crash address format.
      self.update_state_on_match(
          LINUX_GDB_CRASH_ADDRESS_REGEX, line, state, address_from_group=1)

      # Platform specific: Linux gdb crash address format no registers
      self.update_state_on_match(
          LINUX_GDB_CRASH_ADDRESS_NO_REGISTERS_REGEX,
          line,
          state,
          address_from_group=1)

      # Platform specific: Mac gdb style crash address format.
      self.update_state_on_match(
          MAC_GDB_CRASH_ADDRESS_REGEX, line, state, address_from_group=1)

      # Platform specific: Windows cdb style crash type and address format.
      if self.update_state_on_match(
          WINDOWS_CDB_CRASH_TYPE_ADDRESS_REGEX,
          line,
          state,
          type_from_group=1,
          address_from_group=2,
          type_filter=lambda s: s.upper()):
        # Use a consistent format for CDB stacks.
        state.crash_address = '0x%s' % state.crash_address

      # MemorySanitizer / ThreadSanitizer crashes.
      # Make sure to skip the end marker |SUMMARY:|.
      if ' suppressions' not in line and ' warnings' not in line:
        self.update_state_on_match(
            MSAN_TSAN_REGEX,
            line,
            state,
            reset=True,
            type_from_group=2,
            type_filter=lambda s: s.capitalize())

      # LSan can report multiple stacks, so do not clear existing state unless
      # this is a report for an indirect leak. Direct leaks are higher priority.
      if not state.crash_type or state.crash_type == 'Indirect-leak':
        self.update_state_on_match(
            LSAN_DIRECT_LEAK_REGEX,
            line,
            state,
            new_type='Direct-leak',
            reset=True)

      # It's possible that we have a cycle that causes us to only detect
      # indirect leaks, and LSan reports them after any direct leaks. If an
      # indirect leak accompanies a direct leak, we don't care about it.
      if not state.crash_type:
        self.update_state_on_match(
            LSAN_INDIRECT_LEAK_REGEX,
            line,
            state,
            new_type='Indirect-leak',
            reset=True)

      # UndefinedBehavior Sanitizer VPTR (bad-cast) crash.
      if not state.crash_type and self.include_ubsan:
        ubsan_vptr_match = self.update_state_on_match(
            UBSAN_VPTR_REGEX,
            line,
            state,
            new_type='Bad-cast',
            new_frame_count=0,
            address_from_group=3)
        if ubsan_vptr_match:
          state.crash_state = 'Bad-cast to %s' % (
              ubsan_vptr_match.group(4)).strip("'")
          state.found_bad_cast_crash_end_marker = False

      # Get source type information for bad-cast.
      if (state.crash_type == 'Bad-cast' and
          not state.found_bad_cast_crash_end_marker):
        downcast_match = UBSAN_VPTR_INVALID_DOWNCAST_REGEX.match(line)
        if not downcast_match:
          downcast_match = CFI_INVALID_DOWNCAST_REGEX.match(line)
        if downcast_match:
          state.crash_state += ' from %s' % (downcast_match.group(1)).strip("'")
          state.found_bad_cast_crash_end_marker = True

        if (UBSAN_VPTR_INVALID_VPTR_REGEX.match(line) or
            CFI_INVALID_VPTR_REGEX.match(line)):
          state.crash_state += ' from invalid vptr'
          state.found_bad_cast_crash_end_marker = True

        if CFI_FUNC_DEFINED_HERE_REGEX.match(line):
          state.found_bad_cast_crash_end_marker = True

        # Ubsan's -fsanitize=vptr crash extra info for member access.
        invalid_offset_match = UBSAN_VPTR_INVALID_OFFSET_REGEX.match(line)
        if invalid_offset_match:
          state.crash_state += (' from base class subobject at offset %s' %
                                invalid_offset_match.group(1))
          state.found_bad_cast_crash_end_marker = True

        if state.found_bad_cast_crash_end_marker:
          state.crash_state += '\n'
          state.frame_count += 1

      # CFI bad-cast crash.
      if not state.crash_type:
        cfi_bad_cast_match = self.update_state_on_match(
            CFI_ERROR_REGEX,
            line,
            state,
            new_type='Bad-cast',
            new_frame_count=0)
        if cfi_bad_cast_match:
          state.crash_state = 'Bad-cast to %s' % (
              cfi_bad_cast_match.group(2).strip("'"))
          if cfi_bad_cast_match.group(4):
            state.crash_address = cfi_bad_cast_match.group(4)
          state.found_bad_cast_crash_end_marker = False

      # CFI bad-cast crash without extra debugging information.
      if not state.crash_type:
        self.update_state_on_match(
            CFI_NODEBUG_ERROR_MARKER_REGEX,
            line,
            state,
            new_type='Bad-cast',
            new_frame_count=0)

      # Other UndefinedBehavior Sanitizer crash.
      ubsan_runtime_match = UBSAN_RUNTIME_ERROR_REGEX.match(line)
      if ubsan_runtime_match and not state.crash_type and self.include_ubsan:
        reason = ubsan_runtime_match.group(2)
        state.crash_type = 'UNKNOWN'

        for ubsan_crash_regex, ubsan_crash_type in UBSAN_CRASH_TYPES_MAP:
          if self.update_state_on_match(
              ubsan_crash_regex, reason, state, new_type=ubsan_crash_type):
            break

        if state.crash_type == 'UNKNOWN':
          logs.log_error(
              'Unknown UBSan crash type: {reason}'.format(reason=reason))

        state.crash_address = ''
        state.crash_state = ''
        state.frame_count = 0

      # AddressSanitizer for memory overlap crash.
      self.update_state_on_match(
          ASAN_MEMCPY_OVERLAP_REGEX,
          line,
          state,
          new_type='Memcpy-param-overlap',
          reset=True,
          address_from_group=2)

      # Golang stacktraces.
      if state.is_golang:
        for golang_crash_regex, golang_crash_type in GOLANG_CRASH_TYPES_MAP:
          if self.update_state_on_match(
              golang_crash_regex, line, state, new_type=golang_crash_type):
            state.found_golang_crash = True
            state.crash_state = ''
            state.frame_count = 0
            continue

      # Python stacktraces.
      if state.is_python:
        for python_crash_regex, python_crash_type in PYTHON_CRASH_TYPES_MAP:
          if self.update_state_on_match(
              python_crash_regex, line, state, new_type=python_crash_type):
            state.found_python_crash = True
            state.crash_state = ''
            state.frame_count = 0
            continue

      # Sanitizer SEGV crashes.
      segv_match = SAN_SEGV_REGEX.match(line)
      if segv_match:
        temp_crash_address = segv_match.group(3)
        if 'ASSERT' in state.crash_type:
          # We usually use the last crash from the stacktrace for the crash
          # state, but when we see an UNKNOWN crash triggered by an ASSERT, we
          # don't want to overwrite the type, state, and address.
          int_crash_address = crash_analyzer.address_to_integer(
              temp_crash_address)
          if (crash_analyzer.is_assert_crash_address(int_crash_address) or
              SAN_SIGNAL_REGEX.match(stacktrace)):
            continue

        state.crash_type = 'UNKNOWN'
        state.crash_address = temp_crash_address
        state.crash_state = ''
        state.frame_count = 0
        continue

      # AddressSanitizer free on non malloc()-ed address.
      if self.update_state_on_match(
          ASAN_INVALID_FREE_REGEX,
          line,
          state,
          new_type='Invalid-free',
          reset=True,
          address_from_group=1):
        continue

      # AddressSanitizer double free crash.
      if self.update_state_on_match(
          ASAN_DOUBLE_FREE_REGEX,
          line,
          state,
          new_type='Heap-double-free',
          reset=True,
          address_from_group=3):
        continue

      # Sanitizer floating point exception.
      if self.update_state_on_match(
          SAN_FPE_REGEX,
          line,
          state,
          new_type='Floating-point-exception',
          reset=True):
        continue

      # Sanitizer regular crash (includes ills, abrt, etc).
      # However if we have a Go, Python, or Kernel crash, don't replace
      # this with sanitizer signal handler failure.
      if (not state.crash_type.startswith('Kernel failure') and
          not state.found_golang_crash and not state.found_python_crash):
        self.update_state_on_match(
            SAN_ADDR_REGEX,
            line,
            state,
            type_from_group=2,
            address_from_group=4,
            reset=True,
            type_filter=fix_sanitizer_crash_type)

      # Overwrite Unknown-crash type with more generic UNKNOWN type.
      if state.crash_type == 'Unknown-crash':
        state.crash_type = 'UNKNOWN'

      # Sanitizer SEGV type for unknown crashes.
      segv_type_match = SAN_SEGV_CRASH_TYPE_REGEX.match(line)
      if segv_type_match and state.crash_type == 'UNKNOWN':
        segv_type = segv_type_match.group(1)
        if segv_type != 'UNKNOWN':
          state.crash_type += ' ' + segv_type

      # Sanitizer crash type and address format.
      if not state.is_kasan:
        crash_type_and_address_match = self.update_state_on_match(
            SAN_CRASH_TYPE_ADDRESS_REGEX, line, state, address_from_group=3)
        if crash_type_and_address_match and not state.crash_type.startswith(
            'UNKNOWN'):
          state.crash_type += '\n%s %s' % (crash_type_and_address_match.group(
              1).upper(), crash_type_and_address_match.group(2))

      # Android SEGVs.
      # Exclude fatal signal lines from resetting state when we already have
      # one.  Fatal signal lines can often follow the same stack we already
      # processed before. If we process these, we will lose the crash state.
      state_needs_change = (not state.crash_type.startswith('UNKNOWN') or
                            'Fatal signal' not in line)
      if state_needs_change:
        android_segv_match = self.update_state_on_match(
            ANDROID_SEGV_REGEX, line, state, new_type='UNKNOWN', reset=True)
        if android_segv_match:
          state.found_java_exception = False

          # Set the crash address for SEGVs.
          if 'SIGSEGV' in line:
            state.crash_address = android_segv_match.group(1)
            if not state.crash_address.startswith('0x'):
              state.crash_address = '0x%s' % state.crash_address

          # Set process name (if available).
          process_name_match = ANDROID_PROCESS_NAME_REGEX.match(
              android_segv_match.group(2))
          if process_name_match:
            state.process_name = process_name_match.group(1).capitalize()

      # Android SIGABRT handling.
      android_abort_match = self.update_state_on_match(
          ANDROID_ABORT_REGEX,
          line,
          state,
          new_type='CHECK failure',
          new_address='')
      if android_abort_match:
        state.found_java_exception = True
        abort_string = android_abort_match.group(1)
        parts = abort_string.split(' ', 1)
        if len(parts) == 2:
          check_failure_string = parts[1]
          filename_without_fatal = parts[0].replace('FATAL:', '')
          stack_frame = '%s in %s' % (
              fix_check_failure_string(check_failure_string),
              fix_filename_string(filename_without_fatal))
        else:
          stack_frame = fix_check_failure_string(abort_string)
        state.crash_state = stack_frame + '\n'
        state.frame_count = 1

      # Android kernel errors are only checked if this is not a KASan build.
      # Otherwise, we might overwrite the KASan report which contains more
      # useful information.
      if not state.is_kasan:
        self.update_state_on_match(
            ANDROID_KERNEL_ERROR_REGEX,
            line,
            state,
            new_type='Kernel failure',
            reset=True,
            type_from_group=3,
            type_filter=get_fault_description_for_android_kernel)

      # Generic KASan errors.
      if self.update_state_on_match(
          KASAN_CRASH_TYPE_ADDRESS_REGEX,
          line,
          state,
          new_type='Kernel failure',
          type_from_group=1,
          address_from_group=4,
          type_filter=filter_kasan_crash_type):
        state.crash_address = '0x%s' % state.crash_address

      # Generic KASan errors without an address.
      self.update_state_on_match(
          KASAN_CRASH_TYPE_FUNCTION_REGEX,
          line,
          state,
          new_type='Kernel failure',
          type_from_group=1,
          type_filter=filter_kasan_crash_type)

      # KASan GPFs.
      self.update_state_on_match(
          KASAN_GPF_REGEX,
          line,
          state,
          new_type='Kernel failure\nGeneral-protection-fault')

      # For KASan crashes, additional information about a bad access may come
      # from a later line. Update the type and address if this happens.
      update_kasan_crash_details(state, line)

      # Sanitizer tool check failure.
      san_check_match = self.update_state_on_match(
          SAN_CHECK_FAILURE_REGEX,
          line,
          state,
          new_type='Sanitizer CHECK failure',
          new_address='',
          new_frame_count=MAX_CRASH_STATE_FRAMES)
      if san_check_match:
        state.crash_state = '%s\n' % san_check_match.group(1)
        state.process_died = True
        continue

      # Security check failures.
      self.update_state_on_check_failure(
          state, line, SECURITY_CHECK_FAILURE_REGEX, 'Security CHECK failure')
      self.update_state_on_check_failure(
          state, line, SECURITY_DCHECK_FAILURE_REGEX, 'Security DCHECK failure')

      # Timeout/OOM detected by libFuzzer.
      if self.detect_ooms_and_hangs:
        self.update_state_on_match(
            LIBFUZZER_TIMEOUT_REGEX,
            line,
            state,
            new_type='Timeout',
            reset=True)
        self.update_state_on_match(
            OUT_OF_MEMORY_REGEX,
            line,
            state,
            new_type='Out-of-memory',
            reset=True)

      # The following parsing signatures don't lead to crash state overwrites.
      if not state.crash_type:
        # Windows cdb stack overflow.
        self.update_state_on_match(
            WINDOWS_CDB_STACK_OVERFLOW_REGEX,
            line,
            state,
            new_type='Stack-overflow')

        # Windows cdb generic type regex.
        self.update_state_on_match(
            WINDOWS_CDB_CRASH_TYPE_REGEX,
            line,
            state,
            type_from_group=1,
            type_filter=fix_win_cdb_crash_type)

        # Generic ASan regex.
        self.update_state_on_match(
            ASAN_REGEX,
            line,
            state,
            reset=True,
            type_from_group=2,
            type_filter=fix_sanitizer_crash_type)

        # HWASan object address for allocation tail overwritten is on same line
        # as crash type, so add it here.
        self.update_state_on_match(
            HWASAN_ALLOCATION_TAIL_OVERWRITTEN_ADDRESS_REGEX,
            line,
            state,
            address_from_group=1)

        if self.update_state_on_match(
            JAZZER_JAVA_EXCEPTION_REGEX,
            line,
            state,
            new_type='Uncaught exception'):
          state.found_java_exception = True

        # Android fatal exceptions.
        if self.update_state_on_match(
            ANDROID_FATAL_EXCEPTION_REGEX,
            line,
            state,
            new_type='Fatal Exception',
            reset=True):
          state.found_java_exception = True

        # Check failures.
        self.update_state_on_check_failure(state, line, GOOGLE_LOG_FATAL_REGEX,
                                           'Fatal error')
        self.update_state_on_check_failure(
            state, line, CHROME_CHECK_FAILURE_REGEX, 'CHECK failure')
        self.update_state_on_check_failure(
            state, line, GOOGLE_CHECK_FAILURE_REGEX, 'CHECK failure')

        # V8 and Golang fatal errors.
        fatal_error_match = self.update_state_on_match(
            FATAL_ERROR_REGEX, line, state, new_type='Fatal error', reset=True)
        if fatal_error_match:
          state.fatal_error_occurred = True
          state.crash_state = _filter_stack_frame(fatal_error_match.group(1))

        if state.is_golang:
          golang_fatal_error_match = self.update_state_on_match(
              GOLANG_FATAL_ERROR_REGEX,
              line,
              state,
              new_type='Fatal error',
              reset=True)
          if golang_fatal_error_match:
            state.crash_state = golang_fatal_error_match.group(1) + '\n'

        # V8 runtime errors.
        if self.detect_v8_runtime_errors:
          runtime_error_match = (
              self.update_state_on_match(
                  RUNTIME_ERROR_REGEX,
                  line,
                  state,
                  new_type='RUNTIME_ASSERT',
                  reset=True))
          if runtime_error_match:
            state.crash_state = _filter_stack_frame(
                runtime_error_match.group(1))
            state.fatal_error_occurred = True

        # V8 abort errors.
        abort_error_match = self.update_state_on_match(
            V8_ABORT_FAILURE_REGEX, line, state, new_type='ASSERT', reset=True)
        if abort_error_match:
          abort_error = abort_error_match.group(1)
          match = V8_ABORT_METADATA_REGEX.match(abort_error)
          if match:
            abort_error = match.group(1)
            abort_filename = fix_filename_string(match.group(2))
            state.crash_state = '%s\n%s' % (abort_error, abort_filename)
          else:
            state.crash_state = abort_error
          state.frame_count = MAX_CRASH_STATE_FRAMES

        # V8 correctness failure errors.
        self.update_state_on_match(
            V8_CORRECTNESS_FAILURE_REGEX,
            line,
            state,
            new_type='V8 correctness failure',
            reset=True)

        # Generic SEGV handler errors.
        self.update_state_on_match(
            GENERIC_SEGV_HANDLER_REGEX,
            line,
            state,
            new_type='UNKNOWN',
            address_from_group=1,
            address_filter=lambda s: '0x' + s,
            reset=True)

        # Libfuzzer fatal signal errors.
        self.update_state_on_match(
            LIBFUZZER_DEADLY_SIGNAL_REGEX,
            line,
            state,
            new_type='Fatal-signal',
            reset=True)

        # Libfuzzer fuzz target exited errors.
        self.update_state_on_match(
            LIBFUZZER_FUZZ_TARGET_EXITED_REGEX,
            line,
            state,
            new_type='Unexpected-exit',
            reset=True)

        # Libfuzzer fuzz target overwrites const input errors.
        self.update_state_on_match(
            LIBFUZZER_OVERWRITES_CONST_INPUT_REGEX,
            line,
            state,
            new_type='Overwrites-const-input',
            reset=True)

        # Missing library (e.g. a shared library missing in build archive).
        self.update_state_on_match(
            LIBRARY_NOT_FOUND_ANDROID_REGEX,
            line,
            state,
            new_type='Missing-library',
            state_from_group=2,
            reset=True)
        self.update_state_on_match(
            LIBRARY_NOT_FOUND_LINUX_REGEX,
            line,
            state,
            new_type='Missing-library',
            state_from_group=1,
            reset=True)

      if state.fatal_error_occurred:
        error_line_match = self.update_state_on_match(
            FATAL_ERROR_LINE_REGEX, line, state, new_type='Fatal error')
        if not error_line_match and self.detect_v8_runtime_errors:
          error_line_match = self.update_state_on_match(
              RUNTIME_ERROR_LINE_REGEX, line, state, new_type='RUNTIME_ASSERT')

        if error_line_match:
          state.check_failure_source_file = fix_filename_string(
              error_line_match.group(1))
          state.crash_state = '%s\n' % state.check_failure_source_file
          continue

        # Generic fatal errors should be replaced by (D)CHECK failures.
        check_failure_match = self.update_state_on_match(
            FATAL_ERROR_DCHECK_FAILURE,
            line,
            state,
            new_type='DCHECK failure',
            reset=True)

        if not check_failure_match:
          check_failure_match = self.update_state_on_match(
              FATAL_ERROR_CHECK_FAILURE,
              line,
              state,
              new_type='CHECK failure',
              reset=True)

        if not check_failure_match:
          check_failure_match = self.update_state_on_match(
              FATAL_ERROR_GENERIC_FAILURE, line, state, reset=True)

        if check_failure_match and check_failure_match.group(2).strip():
          failure_string = fix_check_failure_string(
              check_failure_match.group(2))
          if state.check_failure_source_file:
            state.crash_state = '%s in %s\n' % (failure_string,
                                                state.check_failure_source_file)
          else:
            state.crash_state = '%s\n' % failure_string
          state.frame_count = 1

        new_state = None
        if state.check_failure_source_file:
          new_state = '%s\n' % state.check_failure_source_file
        self.update_state_on_match(
            FATAL_ERROR_UNREACHABLE,
            line,
            state,
            new_state=new_state,
            new_type='Unreachable code',
            reset=True)

      # Check cases with unusual stack start markers.
      self.update_state_on_match(
          WINDOWS_CDB_STACK_START_REGEX,
          line,
          state,
          new_state='',
          new_frame_count=0)

      # Stack frame parsing signatures.
      # Don't allow more stack frames if a certain stop marker is seen.
      if (state.crash_state and
          utils.sub_string_exists_in(STATE_STOP_MARKERS, line)):
        state.frame_count = MAX_CRASH_STATE_FRAMES
        continue

      # No stack frame parsing required until we get a crash type.
      if not state.crash_type:
        continue

      # For JNI errors, don't use stack frames for crash state from art/runtime,
      # since it helps to do testcase de-duplication.
      if JNI_ERROR_STRING in state.crash_state and '/art/runtime/' in line:
        continue

      # Platform specific: Windows cdb style stack frame.
      if self.add_frame_on_match(
          WINDOWS_CDB_STACK_FRAME_REGEX,
          line,
          state,
          group=4,
          frame_spec=WINDOWS_CDB_STACK_FRAME_SPEC):
        continue

      # Platform specific: Linux and mac gdb, ASAN, MSAN, UBSAN style
      # stack frame. Try the regex with symbols first i.e. with
      # addresses and function names.
      if self.add_frame_on_match(
          SAN_STACK_FRAME_REGEX,
          line,
          state,
          group=3,
          frame_spec=SAN_STACK_FRAME_SPEC,
          frame_override_func=llvm_test_one_input_override):
        continue

      # Chrome symbolized stack frame regex.
      if self.add_frame_on_match(
          CHROME_STACK_FRAME_REGEX,
          line,
          state,
          group=4,
          frame_spec=CHROME_STACK_FRAME_SPEC):
        continue

      # Chrome symbolized stack frame regex (Mac only).
      if self.add_frame_on_match(
          CHROME_MAC_STACK_FRAME_REGEX,
          line,
          state,
          group=6,
          demangle=True,
          frame_spec=CHROME_MAC_STACK_FRAME_SPEC):
        continue

      # Chrome symbolized stack frame regex (Windows only).
      if self.add_frame_on_match(
          CHROME_WIN_STACK_FRAME_REGEX,
          line,
          state,
          group=1,
          frame_spec=CHROME_WIN_STACK_FRAME_SPEC):
        continue

      if state.found_java_exception:
        if (state.crash_type in [
            'CHECK failure', 'Fatal Exception', 'Uncaught exception'
        ]):
          self.add_frame_on_match(
              JAVA_EXCEPTION_CRASH_STATE_REGEX, line, state, group=1)
          continue

      # Android kernel stack frame.
      android_kernel_match = self.add_frame_on_match(
          ANDROID_KERNEL_STACK_FRAME_REGEX, line, state, group=3)
      if android_kernel_match:
        state.found_android_kernel_crash = True

        # Update address from the first stack frame unless we already have
        # more detailed information from KASan.
        if state.frame_count == 1 and not state.is_kasan:
          state.crash_address = '0x%s' % android_kernel_match.group(1)
        continue

      # V8 correctness fuzzer metadata.
      if self.add_frame_on_match(
          V8_CORRECTNESS_METADATA_REGEX,
          line,
          state,
          group=1,
          frame_filter=lambda s: s):
        continue

      # Golang stack frames.
      if state.is_golang and self.add_frame_on_match(
          GOLANG_STACK_FRAME_FUNCTION_REGEX,
          line,
          state,
          group=1,
          frame_filter=lambda s: s.split('/')[-1]):
        continue

      # Python stack frames.
      if state.is_python and self.add_frame_on_match(
          PYTHON_STACK_FRAME_FUNCTION_REGEX, line, state, group=3):
        continue

    # Detect cycles in stack overflow bugs and update crash state.
    update_crash_state_for_stack_overflow_if_needed(state)

    # Convert crash parameters into a generic format regardless of the tool
    # used.
    self.filter_crash_parameters(state)

    return state


def filter_addresses_and_numbers(stack_frame):
  """Return a normalized string without unique addresses and numbers."""
  # Remove offset part from end of every line.
  result = re.sub(r'\+0x[0-9a-fA-F]+\n', '\n', stack_frame, re.DOTALL)

  # Replace sections that appear to be addresses with the string "ADDRESS".
  address_expression = r'0x[a-fA-F0-9]{4,}[U]*'
  address_replacement = r'ADDRESS'
  result = re.sub(address_expression, address_replacement, result)

  # Replace sections that appear to be numbers with the string "NUMBER".
  # Cases that we are avoiding:
  # - source.cc:1234
  # - libsomething-1.0.so (to avoid things like NUMBERso in replacements)
  number_expression = r'(^|[^:0-9.])[0-9.]{4,}($|[^A-Za-z0-9.])'
  number_replacement = r'\1NUMBER\2'
  return re.sub(number_expression, number_replacement, result)


def should_ignore_line_for_crash_processing(line, state):
  """Check to see if a line should be displayed in a report, but ignored when
     processing crashes."""
  # If we detected that the process had died, we won't use any further stack
  # frames to make decision on crash parameters.
  if state.process_died:
    return True

  # Ignore console information messages, as they are not relevant to crash
  # parameters parsing.
  if ':INFO:CONSOLE' in line:
    return True

  # Ignore summary lines.
  if 'SUMMARY:' in line:
    return True

  # Ignore warnings from ASan, but not other sanitizer tools.
  if 'WARNING: AddressSanitizer' in line:
    return True

  # Exclusion for mprotect warning on address 0x00010000. This is a harmless
  # coverage buffer size warning, and is fixed in clang r234602.
  if 'failed to mprotect 0x00010000' in line:
    return True

  # Ignore certain lines printed by dump render tree.
  if 'text run at (' in line:
    return True

  # Ignore this unneeded JNI abort error message since it will be followed by
  # the needed stacktrace later.
  if 'Please include Java exception stack in crash report' in line:
    return True

  # Ignore DEADLYSIGNAL lines from sanitizers.
  if SAN_DEADLYSIGNAL_REGEX.match(line):
    return True

  return False


def fix_sanitizer_crash_type(crash_type):
  """Ensure that Sanitizer crashes use generic formats."""
  # General normalization.
  crash_type = crash_type.lower().replace('_', '-').capitalize()

  # Use more generic types for certain Sanitizer ones.
  crash_type = crash_type.replace('Int-divide-by-zero', 'Divide-by-zero')

  return crash_type


def fix_win_cdb_crash_type(crash_type):
  """Convert a Windows CDB crash type into ASAN like format."""
  # Strip application verifier string from crash type suffix.
  crash_type = utils.strip_from_right(crash_type, '_AVRF')

  # Standardize crash type with lowercase, hyphens and capitalization.
  crash_type = crash_type.replace('_', '-').lower().capitalize()

  # Change crash type to other common types.
  crash_type = crash_type.replace('Status-integer-overflow', 'Integer-overflow')
  crash_type = crash_type.replace('Status-integer-divide-by-zero',
                                  'Divide-by-zero')
  return crash_type


def fix_check_failure_string(failure_string):
  """Cleanup values that should not be included in CHECK failure strings."""
  # Remove |CHECK_FAILURE_PATTERN| from start of failure string.
  failure_string = utils.strip_from_left(failure_string, CHECK_FAILURE_PATTERN)

  # Handle cases like "CHECK_EQ( (unsigned)ptr[0],1u) failed: 25 vs. 1".
  # This only happens on Android, where we cannot strip the
  # CHECK_FAILURE_PATTERN, so we looked for "failed:" as preceding string.
  failure_string = re.sub(r'(?<=failed): .*\svs\.\s.*$', r'', failure_string)

  # Handle cases like "len > 0 (-1 vs. 0)".
  failure_string = re.sub(r' \(.*\s+vs\.\s+.*', r'', failure_string)

  # Handle cases like ": '....'", '= "..."', etc.
  failure_string = re.sub(r'\s*[:=]\s*([\'"]).*\1$', r'', failure_string)

  # Strip unneeded chars at end.
  return failure_string.strip(' .\'"[]')


def fix_filename_string(filename_string):
  """Fix filename string to remove line number, path and other invalid chars."""
  # Remove invalid chars at ends first.
  filename_string = filename_string.strip(' .\'"[]')

  # Remove the source line number information.
  filename_string = filename_string.split(':')[0].split('(')[0]

  # Replace backslashes with forward slashes for platform consistency.
  filename_string = filename_string.replace('\\', '/')

  # Remove the path information.
  filename_string = os.path.basename(filename_string)

  return filename_string


def get_fault_description_for_android_kernel(code):
  """Return human readable fault description based on numeric FSR value."""
  # Convert code from string to number.
  try:
    code = int(code, 16)
  except:
    return 'BUG'

  # Figure out where is out-of-bounds read or write.
  if code & 0x800 == 0:
    fault = 'READ'
  else:
    fault = 'WRITE'
  fault += ' '

  # The full status code is bits 12, 10, and 0-3, but we're ignoring 12 and 10.
  status = code & 0b1111
  try:
    fault += ANDROID_KERNEL_STATUS_TO_STRING[status]
  except KeyError:
    fault += 'Unknown'

  fault += ' (%s)' % str(code)
  return 'Kernel failure\n' + fault


def filter_kasan_crash_type(crash_type):
  """Filter a KASan crash type."""
  return 'Kernel failure\n%s' % crash_type.replace(' ', '-').capitalize()


def update_crash_state_for_stack_overflow_if_needed(state):
  """For stack-overflow bugs, updates crash state based on cycle detected."""
  if state.crash_type != 'Stack-overflow':
    return

  num_frames = len(state.raw_frames)
  for frame_index in range(num_frames):
    for cycle_length in range(1, MAX_CYCLE_LENGTH + 1):
      # Create frame potential cycles of a given length starting from
      # |frame_index|.
      frame_potential_cycles = []
      end_reached = False
      for i in range(0, REPEATED_CYCLE_COUNT):
        start_index = frame_index + i * cycle_length
        end_index = frame_index + (i + 1) * cycle_length
        if end_index >= num_frames:
          end_reached = True
          break

        frame_potential_cycles.append(state.raw_frames[start_index:end_index])

      if end_reached:
        # Reached end while trying to find cycle, skip iteration.
        continue

      # Check if all the potential_cycles are equal. If yes, we found a cycle.
      potential_cycles_are_equal = all(
          frame_potential_cycle == frame_potential_cycles[0]
          for frame_potential_cycle in frame_potential_cycles)

      # Update crash state based on cycle detected.
      if potential_cycles_are_equal:
        state.crash_state = '\n'.join(
            frame_potential_cycles[0][:MAX_CRASH_STATE_FRAMES]) + '\n'
        return


def llvm_test_one_input_override(frame, frame_struct):
  """Override frame matching for LLVMFuzzerTestOneInput frames."""
  if not frame.startswith('LLVMFuzzerTestOneInput'):
    return frame

  if frame_struct and frame_struct.filename:
    # Use the filename as the frame instead.
    return frame.replace(
        'LLVMFuzzerTestOneInput',
        os.path.basename(frame_struct.filename.replace('\\', '/')))

  return frame


def reverse_python_stacktrace(stacktrace):
  """Extract a Python stacktrace.
  Python stacktraces are a bit special: they are reversed,
  and followed by a sanitizer one, so we need to extract them, reverse them,
  and put their "title" back on top."""
  python_stacktrace_split = []
  in_python_stacktrace = False

  for line in stacktrace:
    # Locate the beginning of the python stacktrace.
    if in_python_stacktrace is False:
      for regex, _ in PYTHON_CRASH_TYPES_MAP:
        if regex.match(line):
          in_python_stacktrace = True
          python_stacktrace_split = [line]  # Add the "title" of the stacktrace
          break
    else:
      if '=========' in line:  # Locate beginning of the sanitizer stacktrace
        break
      python_stacktrace_split.insert(1, line)

  return python_stacktrace_split


def update_kasan_crash_details(state, line):
  """For KASan crashes, additional information about a bad access may exist."""
  if state.crash_type.startswith('Kernel failure'):
    kasan_access_match = KASAN_ACCESS_TYPE_ADDRESS_REGEX.match(line)
    if kasan_access_match:
      if not state.crash_address:
        state.crash_address = '0x%s' % kasan_access_match.group(4)
    else:
      kasan_access_match = KASAN_ACCESS_TYPE_REGEX.match(line)

    if kasan_access_match:
      state.crash_type += '\n%s %s' % (kasan_access_match.group(1).upper(),
                                       kasan_access_match.group(2))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzing engine definition for generic blackbox fuzzers."""

import os

from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_analyzer
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import shell
from clusterfuzz.fuzz import engine

TESTCASE_PREFIX = 'fuzz-'
OUTPUT_PREFIX = 'output-'


def _get_arguments(app_path, app_args):
  """Get arguments shared between multiple run types."""
  return [f'--app_path={app_path}', f'--app_args={app_args}']


def _run_with_interpreter_if_needed(fuzzer_path, args, max_time):
  """Execute the fuzzer script with an interpreter, or invoke it directly."""
  interpreter = shell.get_interpreter(fuzzer_path)
  if interpreter:
    executable = interpreter
    args.insert(0, fuzzer_path)
  else:
    executable = fuzzer_path

  runner = new_process.UnicodeProcessRunner(executable)
  return runner.run_and_wait(timeout=max_time, additional_args=args)


class BlackboxEngine(engine.Engine):
  """Generic blackbox fuzzer engine implementation.

  Blackbox fuzzers are scripts which accept the following arguments:
    --app_path: The path to a binary which will be invoked when testing.
    --app_args: Arguments to pass to the test binary.
    --input_dir: An optional corpus containing test data.
    --output_dir: If specified, write generated test cases to this directory
                  along with their output. This is specified during fuzzing but
                  not reproduction. Only the crashes must be written, but
                  crashes will be identified automatically if the fuzzer does
                  not detect them itself. Tests should have the prefix "fuzz-",
                  and their output should have the same name as the test but
                  with the prefix "output-" instead of "fuzz-".
    --testcase_path: If specified, a path to a test case to reproduce. Output
                     should be written directly to stdout or stderr.
  """

  @property
  def name(self):
    return 'blackbox'

  def prepare(self, corpus_dir, target_path, build_dir):  # pylint: disable=unused-argument
    """Prepare for a fuzzing session by generating options.

    Though blackbox fuzzers follow the engine interface, they must be launched
    in a different manner from most other engine fuzzers. Instead of running a
    target directly, these fuzzers tend to be wrapper scripts which generate
    test cases and pass them to a binary that is managed by the infrastructure.

    Args:
      corpus_dir: The main corpus directory.
      target_path: Path to the fuzzer script or binary.
      build_dir: Path to the build directory.
    Returns:
      A FuzzOptions object.
    """
    return engine.FuzzOptions(corpus_dir, [], {})

  # TODO(mbarbella): As implemented, this will not work for untrusted workers.
  # We would need to copy fuzzer binaries to workers.
  def fuzz(self, target_path, options, reproducers_dir, max_time):
    """Run a fuzzing session.
    Args:
      target_path: Path to the fuzzer script or binary.
      options: The FuzzOptions object returned by prepare().
      reproducers_dir: The directory to put reproducers in when crashes
          are found.
      max_time: Maximum allowed time for the fuzzing to run.
   Returns:
      A FuzzResult object.
    """
    # For blackbox fuzzers, |target_path| supplies the path to the fuzzer script
    # rather than a target in the build archive.
    fuzzer_path = target_path
    os.chmod(fuzzer_path, 0o775)

    app_path = environment.get_value('APP_PATH')
    app_args = testcase_manager.get_command_line_for_application(
        get_arguments_only=True).strip()
    corpus_dir = options.corpus_dir
    command_line_args = _get_arguments(app_path, app_args)
    command_line_args.append(f'--input_dir={corpus_dir}')

    result = _run_with_interpreter_if_needed(fuzzer_path, command_line_args,
                                             max_time)
    crashes = []
    for testcase_path in os.listdir(reproducers_dir):
      if not testcase_path.startswith(TESTCASE_PREFIX):
        continue

      output_path = OUTPUT_PREFIX + testcase_path[len(TESTCASE_PREFIX):]
      absolute_output_path = os.path.join(reproducers_dir, output_path)

      # If no output was written for a test case, skip it.
      if not os.path.exists(absolute_output_path):
        continue

      with open(absolute_output_path, 'r', errors='replace') as handle:
        output = handle.read()

      # Filter obviously non-crashing test cases. Crashes still follow the
      # normal flow in fuzz task to ensure that the state should not be ignored
      # for other reasons, but we don't want to log every test case for the
      # fuzzers that don't do their own crash processing.
      state = stack_analyzer.get_crash_data(output)
      if not state.crash_type:
        continue

      full_testcase_path = os.path.join(reproducers_dir, testcase_path)
      crash = engine.Crash(full_testcase_path, output, options.arguments,
                           int(result.time_executed))
      crashes.append(crash)

    # TODO(mbarbella): Support stats.
    stats = {}

    return engine.FuzzResult(result.output, result.command, crashes, stats,
                             result.time_executed)

  def reproduce(self, target_path, input_path, arguments, max_time):
    """Reproduce a crash given an input.
    Args:
      target_path: Path to the fuzzer script or binary.
      input_path: Path to the reproducer input.
      arguments: Additional arguments needed for reproduction.
      max_time: Maximum allowed time for the reproduction.
    Returns:
      A ReproduceResult.
    """
    del arguments
    # For blackbox fuzzers, |target_path| supplies the path to the fuzzer script
    # rather than a target in the build archive.
    fuzzer_path = target_path
    os.chmod(fuzzer_path, 0o775)

    app_path = environment.get_value('APP_PATH')
    app_args = testcase_manager.get_command_line_for_application(
        get_arguments_only=True).strip()

    args = _get_arguments(app_path, app_args)
    args.append(f'--testcase_path={input_path}')

    result = _run_with_interpreter_if_needed(fuzzer_path, args, max_time)
    return engine.ReproduceResult(result.command, result.return_code,
                                  result.time_executed, result.output)

  def minimize_corpus(self, target_path, arguments, input_dirs, output_dir,
                      reproducers_dir, max_time):
    """Run corpus minimization.
    Args:
      target_path: Path to the fuzzer script or binary.
      arguments: Additional arguments needed for corpus minimization.
      input_dirs: Input corpora.
      output_dir: Output directory to place minimized corpus.
      reproducers_dir: The directory to put reproducers in when crashes are
          found.
      max_time: Maximum allowed time for the minimization.
    Returns:
      A FuzzResult object.
    """
    # Blackbox fuzzers follow something closer to the traditional model, and as
    # of now do not rely on corpora.
    raise NotImplementedError(
        'Corpus minimization is not supported for blackbox fuzzers.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Security Severity analysis."""

# TODO(ochang): Support other sanitizers.

import re

from clusterfuzz._internal.datastore.data_types import MISSING_VALUE_STRING
from clusterfuzz._internal.datastore.data_types import SecuritySeverity
from clusterfuzz._internal.system import environment

# These should be generic within ClusterFuzz.
LOW_SEVERITY_CRASH_TYPES = []
MEDIUM_SEVERITY_CRASH_TYPES = [
    'Container-overflow', 'Heap-buffer-overflow',
    'Incorrect-function-pointer-type', 'Index-out-of-bounds',
    'Memcpy-param-overlap', 'Non-positive-vla-bound-value', 'Object-size',
    'Stack-buffer-overflow', 'UNKNOWN', 'Use-of-uninitialized-value'
]
HIGH_SEVERITY_CRASH_TYPES = [
    'Bad-cast', 'Heap-double-free', 'Heap-use-after-free',
    'Security DCHECK failure', 'Use-after-poison'
]

SEVERITY_ORDER = [
    SecuritySeverity.LOW, SecuritySeverity.MEDIUM, SecuritySeverity.HIGH,
    SecuritySeverity.CRITICAL
]


def _modify_severity(severity,
                     delta,
                     min_severity=SecuritySeverity.LOW,
                     max_severity=SecuritySeverity.CRITICAL):
  """Increase/decrease the given |severity| by |delta|."""
  min_index = SEVERITY_ORDER.index(min_severity)
  max_index = SEVERITY_ORDER.index(max_severity)
  assert min_index != -1 and max_index != -1

  severity_index = SEVERITY_ORDER.index(severity)
  assert severity_index != -1

  max_index = min(len(SEVERITY_ORDER) - 1, max_index)

  severity_index += delta
  severity_index = min(severity_index, max_index)
  severity_index = max(severity_index, min_index)

  return SEVERITY_ORDER[severity_index]


def get_analyzer(name):
  """Return an analyzer for the given |name|."""
  if name == 'sanitizer_generic':
    return SeverityAnalyzerSanitizer()
  if name == 'sanitizer_chrome':
    return SeverityAnalyzerSanitizerChrome(is_compromised_renderer=False)
  if name == 'sanitizer_chrome_compromised_renderer':
    return SeverityAnalyzerSanitizerChrome(is_compromised_renderer=True)
  return None


def get_security_severity(crash_type, crash_output, job_name,
                          requires_gestures):
  """Convenience function to get the security severity of a crash."""
  analyzer = None
  severity_analyzer_name = environment.get_value('SECURITY_SEVERITY_ANALYZER')

  if severity_analyzer_name:
    analyzer = get_analyzer(severity_analyzer_name)
  else:
    is_chrome = 'chrome' in job_name or 'content_shell' in job_name
    is_sanitizer = ('_asan' in job_name or '_cfi' in job_name or
                    '_lsan' in job_name or '_msan' in job_name or
                    '_tsan' in job_name or '_ubsan' in job_name)

    if is_sanitizer:
      if is_chrome:
        analyzer = get_analyzer('sanitizer_chrome')
      else:
        analyzer = get_analyzer('sanitizer_generic')

  if not analyzer:
    return None

  return analyzer.analyze(crash_type, crash_output, requires_gestures)


class SeverityAnalyzerSanitizer(object):
  """Generic ASan severity analyzer."""

  def analyze(self, crash_type, crash_output, requires_gestures):
    """Return a security severity based on the ASan crash output."""

    manual_severity_match = re.search(
        r'FuzzerSecurityIssue(Critical|High|Medium|Low)', crash_output)
    if manual_severity_match:
      manual_severity = manual_severity_match.group(1)
      return string_to_severity(manual_severity)

    crash_category = crash_type.split()[0]
    if crash_category in HIGH_SEVERITY_CRASH_TYPES:
      severity = SecuritySeverity.HIGH
    elif crash_category in MEDIUM_SEVERITY_CRASH_TYPES:
      severity = SecuritySeverity.MEDIUM
    elif crash_category in LOW_SEVERITY_CRASH_TYPES:
      severity = SecuritySeverity.LOW
    else:
      return None

    if requires_gestures:
      severity = _modify_severity(severity, -1)

    if 'WRITE' in crash_type:
      severity = _modify_severity(
          severity, 1, max_severity=SecuritySeverity.HIGH)

    # TODO(ochang): Detect really weird stacks, and bump them up to high.
    # TODO(ochang): ASSERTs should be high, but right now we also hit release
    # asserts, and there's no way to tell just from the stacktrace.
    # TODO(ochang): Once ASan has some exploitability analysis, take that into
    # account too to bump up/down the severity.

    return severity


class SeverityAnalyzerSanitizerChrome(SeverityAnalyzerSanitizer):
  """Chrome specific severity analyzer."""

  PROCESS_TYPE_EXCEPTIONS = [
      'Use-of-uninitialized-value',
  ]

  def __init__(self, is_compromised_renderer):
    SeverityAnalyzerSanitizer.__init__(self)
    self.is_compromised_renderer = is_compromised_renderer

  def analyze(self, crash_type, crash_output, requires_gestures):
    """Return a security severity based on the ASan crash output."""
    # Base severity.
    severity = SeverityAnalyzerSanitizer.analyze(self, crash_type, crash_output,
                                                 requires_gestures)
    if severity is None:
      return None

    # Chrome specific severity adjustments.
    if (crash_type not in self.PROCESS_TYPE_EXCEPTIONS and
        not self.is_compromised_renderer and
        self._find_process_type(crash_output) == 'browser'):
      severity = _modify_severity(severity, 1)

    return severity

  @staticmethod
  def _find_process_type(crash_output):
    """Return the process type of the process that crashed."""
    # TODO(ochang): Support Android.
    process_type = None
    # This is best effort, and won't work in cases where we have weird stacks
    # (or V8). Since we only care right now if this is a browser process (where
    # this should be rare from an uncompromised renderer), it shouldn't matter
    # too much.
    main_function_regex = re.compile(r'content::([A-Z][a-z]+)Main\(')

    # As a fallback, search for content/browser file paths for determining the
    # browser process.
    content_browser_regex = re.compile(r'content[/\\]browser')

    for line in crash_output.splitlines():
      if 'content::ContentMain' in line:
        continue

      if content_browser_regex.search(line):
        process_type = 'browser'
        break

      match = main_function_regex.search(line)
      if not match:
        continue

      process_type = match.group(1).lower()
      break

    return process_type


def severity_to_string(severity):
  """Convert a severity value to a human-readable string."""
  severity_map = {
      SecuritySeverity.CRITICAL: 'Critical',
      SecuritySeverity.HIGH: 'High',
      SecuritySeverity.MEDIUM: 'Medium',
      SecuritySeverity.LOW: 'Low',
      SecuritySeverity.MISSING: MISSING_VALUE_STRING,
  }

  return severity_map.get(severity, '')


def string_to_severity(severity):
  """Convert a string value to a severity value."""
  severity_map = {
      'critical': SecuritySeverity.CRITICAL,
      'high': SecuritySeverity.HIGH,
      'medium': SecuritySeverity.MEDIUM,
      'low': SecuritySeverity.LOW,
  }

  if severity.lower() in severity_map:
    return severity_map[severity.lower()]

  return SecuritySeverity.MISSING



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reproduce tool tests."""
# pylint: disable=protected-access
import os
import tempfile
import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from clusterfuzz._internal.tests.test_libs.reproduce_tool_fakes import \
    FakeConfig
from clusterfuzz._internal.tests.test_libs.reproduce_tool_fakes import \
    FakeResponse
from local.butler import reproduce
from local.butler.reproduce_tool import errors


def _fake_get_echo_testcase(*_):
  """Fake test case output intended to run "echo -n"."""
  testcase_map = {
      'crash_state': 'state',
      'security_flag': False,
      'gestures': [],
      'flaky_stack': False,
      'job_type': 'test_job',
      'redzone': 32,
      'disable_ubsan': False,
      'additional_metadata': '{}',
      'fuzzer_name': 'fuzzer',
      'job_definition': 'APP_NAME = echo\nAPP_ARGS = -n\n',
      'overridden_fuzzer_name': 'fuzzer',
      'platform': environment.platform().lower(),
      'minimized_arguments': '',
      'window_argument': '',
      'timeout_multiplier': 1.0,
      'serialized_fuzz_target': None,
      'one_time_crasher_flag': False,
  }

  return reproduce.SerializedTestcase(testcase_map)


def _fake_get_libfuzzer_testcase(*_):
  """Fake test case output intended to run "echo -n"."""
  testcase_map = {
      'crash_state': 'state',
      'security_flag': False,
      'gestures': [],
      'flaky_stack': False,
      'job_type': 'test_job',
      'redzone': 32,
      'disable_ubsan': False,
      'additional_metadata': '{}',
      'fuzzer_name': 'libFuzzer',
      'job_definition': 'APP_NAME = launcher.py\n',
      'overridden_fuzzer_name': 'libFuzzer_test_fuzzer',
      'platform': environment.platform().lower(),
      'minimized_arguments': '',
      'window_argument': '',
      'timeout_multiplier': 1.0,
      'serialized_fuzz_target': {
          'binary': 'test_fuzzer',
          'engine': 'libFuzzer',
          'project': 'test_project',
      },
      'one_time_crasher_flag': False,
  }

  return reproduce.SerializedTestcase(testcase_map)


@test_utils.reproduce_tool
@test_utils.with_cloud_emulators('datastore')
class ReproduceTest(unittest.TestCase):
  """Tests for the full reproduce tool."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.bot.testcase_manager.engine_reproduce',
        'clusterfuzz._internal.config.local_config.ProjectConfig',
        'clusterfuzz.fuzz.engine.get',
        'local.butler.reproduce._download_testcase',
        'local.butler.reproduce._get_testcase',
        'local.butler.reproduce._setup_x',
        'local.butler.reproduce._verify_target_exists',
        'local.butler.reproduce_tool.config.ReproduceToolConfiguration',
        'local.butler.reproduce_tool.prompts.get_boolean',
        'clusterfuzz._internal.system.process_handler.run_process',
        'clusterfuzz._internal.system.process_handler.'
        'terminate_stale_application_instances',
    ])
    helpers.patch_environ(self)

    self.mock._setup_x.return_value = []
    self.mock.get_boolean.return_value = True
    self.mock._download_testcase.return_value = '/tmp/testcase'
    self.mock.run_process.return_value = (0, 0, '/tmp/testcase')

    self.build_directory = tempfile.mkdtemp()

  def tearDown(self):
    shell.remove_directory(self.build_directory)

  def test_reproduce_with_echo(self):
    """Ensure that the tool can run a job configured to execute "echo"."""
    self.mock.get.return_value = None
    self.mock._get_testcase.side_effect = _fake_get_echo_testcase

    binary_path = os.path.join(self.build_directory, 'echo')
    with open(binary_path, 'w') as f:
      f.write('test')

    crash_retries = 3
    disable_xvfb = False
    verbose = False
    disable_android_setup = False
    application = None
    reproduce._reproduce_crash('https://localhost/testcase-detail/1',
                               self.build_directory, crash_retries,
                               disable_xvfb, verbose, disable_android_setup,
                               application)
    reproduce._cleanup()
    self.mock.run_process.assert_called_with(
        binary_path + ' -n /tmp/testcase',
        current_working_directory=self.build_directory,
        gestures=[],
        timeout=10)

    # The tool does an initial run before running |crash_retries| times.
    self.assertEqual(self.mock.run_process.call_count, crash_retries + 1)

  def test_reproduce_with_libfuzzer(self):
    """Ensure that the tool can run on a libFuzzer target."""
    self.mock.get.return_value = 'fake engine object'
    self.mock._get_testcase.side_effect = _fake_get_libfuzzer_testcase

    crash_retries = 3
    disable_xvfb = False
    verbose = False
    disable_android_setup = False
    application = None
    reproduce._reproduce_crash('https://localhost/testcase-detail/1',
                               self.build_directory, crash_retries,
                               disable_xvfb, verbose, disable_android_setup,
                               application)
    reproduce._cleanup()
    self.mock.engine_reproduce.assert_called_with(
        'fake engine object', 'test_fuzzer', '/tmp/testcase', [], 10)

    # The tool does an initial run before running |crash_retries| times.
    self.assertEqual(self.mock.engine_reproduce.call_count, crash_retries + 1)


@test_utils.reproduce_tool
class DownloadTestcaseTest(unittest.TestCase):
  """Tests for _download_testcase."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.write_data_to_file',
        'local.butler.reproduce_tool.http_utils.request',
        'clusterfuzz._internal.system.archive.unpack',
        'clusterfuzz._internal.system.archive.get_file_list',
    ])
    helpers.patch_environ(self)

    self.config = FakeConfig()

  def test_initial_request_failed(self):
    """Ensure that we bail out if the initial request fails."""
    self.mock.request.return_value = (FakeResponse(500), '')
    testcase = reproduce.SerializedTestcase({})
    with self.assertRaises(errors.ReproduceToolUnrecoverableError):
      reproduce._download_testcase(1, testcase, self.config)

  def test_non_archived_testcase(self):
    """Ensure that we properly unpack non-archived test cases."""
    self.mock.request.return_value = (FakeResponse(200, filename='test.html'),
                                      'html data')
    testcase = reproduce.SerializedTestcase({
        'archive_state': data_types.ArchiveStatus.NONE,
        'minimized_keys': 'key',
    })

    reproduce._download_testcase(1, testcase, self.config)
    self.mock.write_data_to_file.assert_called_once_with(
        'html data',
        os.path.join(reproduce.CONFIG_DIRECTORY, 'current-testcase',
                     'test.html'))

  def test_archived_testcase(self):
    """Ensure that we properly unpack archived test cases."""
    self.mock.request.return_value = (FakeResponse(200, filename='test.zip'),
                                      'zip data')
    self.mock.get_file_list.return_value = ['test.html', 'resource.js']
    testcase = reproduce.SerializedTestcase({
        'archive_state': data_types.ArchiveStatus.ALL,
        'absolute_path': '/path/to/test.html',
        'minimized_keys': 'key',
    })

    current_testcase_directory = os.path.join(reproduce.CONFIG_DIRECTORY,
                                              'current-testcase')
    zip_path = os.path.join(current_testcase_directory, 'test.zip')

    reproduce._download_testcase(1, testcase, self.config)
    self.mock.write_data_to_file.assert_called_once_with('zip data', zip_path)

    self.mock.unpack.assert_called_once_with(zip_path,
                                             current_testcase_directory)

  def test_archive_missing_file(self):
    """Ensure that we raise if the archive is missing an expected file."""
    self.mock.request.return_value = (FakeResponse(200, filename='test.zip'),
                                      'zip data')
    self.mock.get_file_list.return_value = []
    testcase = reproduce.SerializedTestcase({
        'archive_state': data_types.ArchiveStatus.ALL,
        'absolute_path': '/path/to/test.html',
        'minimized_keys': 'key',
    })

    with self.assertRaises(errors.ReproduceToolUnrecoverableError):
      reproduce._download_testcase(1, testcase, self.config)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Methods that are used in multiple places."""
import os
import re

from fabric import api


def get_file_content(file_path):
  """Return content of a file."""
  return open(file_path).read().strip()


def get_host_user_and_ssh_key_path(instance_name, project, zone):
  """Return a tuple of (hostname, username and ssh_key_path)."""
  output = api.local(
      'gcloud compute ssh --project "%s" --zone "%s" %s --dry-run' %
      (project, zone, instance_name),
      capture=True)
  print(output)

  m = re.match('/usr/bin/ssh .*-i ([^ ]+)(?: -o [^ ]+)* ([^ ]+)@([^ ]+)',
               output)
  return (m.group(3), m.group(2), m.group(1))


def get_password():
  """Return password from |PASSWORD_FILE_PATH| environment variable."""
  password_file_path = os.getenv('PASSWORD_FILE_PATH')
  if not password_file_path:
    raise Exception('Please set PASSWORD_FILE_PATH in environment.')

  return get_file_content(password_file_path)
